#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Copyright 2019 Pintaudi Giorgio

""" Apply the high voltage to the MPPCs """

import random
import re

try:
    import bindpyrame
except ImportError:
    bindpyrame = False

from six import string_types
from typing import Optional, Callable, Dict, Tuple


def _silent_simulator(method):
    # type: (Callable) -> Callable
    def call(self, *args, **kwargs):
        # type: (PyrameModule, *str, **str) -> None
        if self.simulate:
            return
        else:
            return method(self, *args, **kwargs)
    return call


def _int_simulator(method):
    # type: (Callable) -> Callable
    def call(self, *args, **kwargs):
        # type: (PyrameModule, *str, **str) -> int
        if self.simulate:
            return random.randint(1, 1000)
        else:
            return method(self, *args, **kwargs)
    return call


def _pyrame_simulator(method):
    # type: (Callable) -> Callable

    def call(self, function, *args):
        # type: (PyrameModule, str, *str) -> Tuple[int, str]
        if self.simulate:
            function_name = re.match('(.+)_' + self.module_name, function).group(1)
            if function_name in self.simulation_dic:
                return self.simulation_dic[function_name]
            else:
                return 1, "ok"
        else:
            return method(self, function, *args)
    return call


class PyrameModule(object):

    def __init__(self, module_name, simulate=False, simulation_dic=None):
        # type: (str, bool, Dict[str, Tuple[int, str]]) -> None
        # Arguments sanity checks
        if not isinstance(module_name, string_types):
            raise ValueError("Pyrame module name is not a string")
        # Attributes initialization
        self._module_name = module_name.lower()
        self._port_name = module_name.upper() + "_PORT"
        self._socket_is_open = False
        self._socket = None
        self._simulate = simulate
        self._simulation_dic = {} if simulation_dic is None else simulation_dic
        # Open socket
        self._get_socket()

    @property
    def module_name(self):
        # type: (...) -> str
        return self._module_name

    @property
    def port_name(self):
        # type: (...) -> str
        return self._port_name

    @property
    def simulate(self):
        # type: (...) -> bool
        return self._simulate

    @property
    def simulation_dic(self):
        # type: (...) -> Dict[str, Tuple[int, str]]
        return self._simulation_dic

    @_int_simulator
    def _open_socket(self):
        # type: (...) -> int
        return bindpyrame.open_socket("localhost", bindpyrame.get_port(self.port_name))

    def _get_socket(self):
        # type: (...) -> None
        self._socket = self._open_socket()
        if not self._socket:
            raise RuntimeError("Error in opening socket")
        self._socket_is_open = True

    @_silent_simulator
    def _close_socket(self):
        # type: (...) -> None
        bindpyrame.close_socket(self._socket)

    def __del__(self):
        # type: (...) -> None
        if self._socket_is_open:
            self._close_socket()
            self._socket_is_open = False

    @_pyrame_simulator
    def pyrame_execcmd(self, function, *argv):
        # type: (str, *str) -> Tuple[int, str]
        return bindpyrame.execcmd(self._socket, function, *argv)

    def execcmd(self, function, *argv):
        # type: (str, *str) -> Tuple[int, str]
        # Argument sanity checks
        arguments = [str(i) for i in argv]
        if not isinstance(function, string_types):
            raise ValueError("Function name is not a string")
        retcode, res = self.pyrame_execcmd(function + "_" + self.module_name, *arguments)
        if retcode == 0:
            raise RuntimeError("module name '%s' : function %s : arguments %s : retcode=%d res=%s"
                               % (self.module_name, function, ','.join(arguments), retcode, res))
        else:
            return retcode, res


#################################################################################################

class PyrameSlowModule(PyrameModule):
    """Class to control a generic Pyrame module"""
    _modules_id_list = []

    def __init__(self, module_name, module_id=None, conf_string=None, simulate=False, simulation_dic=None):
        # type: (str, Optional[str], Optional[str], bool, Optional[Dict[str, Tuple[int, str]]]) -> None
        super(PyrameSlowModule, self).__init__(module_name=module_name, simulate=simulate,
                                               simulation_dic=simulation_dic)
        # Object members
        self._initialized = False
        self._configured = False
        self._module_id = module_id
        # Arguments sanity checks
        if conf_string is not None and not isinstance(conf_string, string_types):
            raise ValueError("Configuration string is not a string")
        # Members initialization
        if not self.module_id:
            self._module_id = "%s_%d" % (self.module_name, random.randint(10000, 99999))
            while self._module_id in self._modules_id_list:
                self._module_id = "%s_%d" % (self.module_name, random.randint(10000, 99999))
        self._modules_id_list.append(self._module_id)
        # Initialize module
        self._initialize(conf_string)
        # Configure module
        self._configure()

    @property
    def module_id(self):
        # type: (...) -> str
        return self._module_id

    def _initialize(self, conf_string=None):
        # type: (Optional[str]) -> None
        try:
            retcode, res = self.execcmd("id_exists")
            if retcode == 1 and res.lower() == "yes":
                self._configure()
                self._deinitialize()
        except RuntimeError:
            pass
        if conf_string is None:
            retcode, res = self.execcmd("init")
        else:
            retcode, res = self.execcmd("init", conf_string)
        if retcode == 0:
            raise RuntimeError("module name '%s' with module ID '%s' : initialize error : retcode=%d res=%s"
                               % (self.module_name, self.module_id, retcode, res))
        self._initialized = True

    def _configure(self):
        # type: (...) -> None
        retcode, res = self.execcmd("config")
        if retcode == 0:
            raise RuntimeError("module name '%s' with module ID '%s' : configure error : retcode=%d res=%s"
                               % (self.module_name, self.module_id, retcode, res))
        self._configured = True

    def _invalidate(self):
        # type: (...) -> None
        retcode, res = self.execcmd("inval")
        if retcode == 0:
            print("module name '%s' with module ID '%s' : invalidate error : retcode=%d res=%s"
                  % (self.module_name, self.module_id, retcode, res))
        self._configured = False

    def _deinitialize(self):
        # type: (...) -> None
        retcode, res = self.execcmd("deinit")
        if retcode == 0:
            print("module name '%s' with module ID '%s' : deinitialize error : retcode=%d res=%s"
                  % (self.module_name, self.module_id, retcode, res))
        self._initialized = False

    def __del__(self):
        # type: (...) -> None
        if self._configured:
            self._invalidate()
        if self._initialized:
            self._deinitialize()
        self._modules_id_list.remove(self.module_id)
        super(PyrameSlowModule, self).__del__()

    def execcmd(self, function, *argv):
        # type: (str, ...) -> Tuple[int, str]
        # Argument sanity checks
        str_argv = []
        for arg in argv:
            str_argv.append(str(arg))
        if not isinstance(function, string_types):
            raise ValueError("Function name is not a string")
        retcode, res = super(PyrameSlowModule, self).execcmd(function, self.module_id, *str_argv)
        if retcode == 0:
            raise RuntimeError("module name '%s' with module ID '%s' : %s error : retcode=%d res=%s"
                               % (self.module_name, self.module_id, function, retcode, res))
        else:
            return retcode, res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio
#

def import_tkinter():
    try:
        # for Python2
        # noinspection PyPep8Naming
        import Tkinter as tkinter
        print("Found Python2 Tkinter module")
    except ImportError:
        # for Python3
        try:
            import tkinter
            print("Found Python3 tkinter module")
        except ImportError:
            print("[ERROR] tkinter could not be found!")
            tkinter = None
    return tkinter


def import_filedialog():
    try:
        # for Python2
        # noinspection PyPep8Naming,PyUnresolvedReferences
        import tkFileDialog as filedialog
        print("Found Python2 tkFileDialog module")
    except ImportError:
        # for Python3
        try:
            # noinspection PyPep8Naming,PyUnresolvedReferences,PyCompatibility
            from tkinter import filedialog
            print("Found Python3 filedialog module")
        except ImportError:
            print("[ERROR] filedialog could not be found!")
            filedialog = None
    return filedialog


def import_messagebox():
    try:
        # for Python2
        # noinspection PyPep8Naming,PyUnresolvedReferences
        import tkMessageBox as messagebox
        print("Found Python2 tkMessageBox module")
    except ImportError:
        # for Python3
        try:
            # noinspection PyPep8Naming,PyUnresolvedReferences,PyCompatibility
            from tkinter import messagebox
            print("Found Python3 messagebox module")
        except ImportError:
            print("[ERROR] messagebox could not be found!")
            messagebox = None
    return messagebox


def import_pygubu():
    try:
        import pygubu
        print("Found Python pygubu module")
    except ImportError:
        print("[ERROR] pygubu could not be found!")
        pygubu = None
    return pygubu



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

import abc
import inspect
import json
import os
import shutil
import time
from collections import OrderedDict
from enum import Enum
from typing import Any, List, Dict, Optional, Union

from bitarray import bitarray
from recordclass import recordclass

import wagascianpy.analysis.analysis as analysis
import wagascianpy.analysis.apply_detector_flags as apply_detector_flags
import wagascianpy.analysis.beam_summary_data as beam_summary_data
import wagascianpy.analysis.sanity_check as sanity_check
import wagascianpy.utils.environment as environment
import wagascianpy.utils.utils as utils
from wagascianpy.utils.acq_config_xml import acqconfigxml_file_finder

try:
    from shutil import SameFileError
except ImportError:
    shutil.SameFileError = OSError

# compatible with Python 2 *and* 3:
ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})
# Chain mutable tuple
Chain = recordclass('Chain', ['link', 'thread'])


class AnalyzerInputType(Enum):
    """
    Type of input of the analyzer. single_run means that the analyzer accepts a single run at a time. multiple_runs
    means that the analyzer combine the information from multiple runs.
    """
    single_run = 1,
    multiple_runs = 2


class AnalyzerThreadingType(Enum):
    """
    Multithreading capabilities of an analyzer. multi_threaded means that the analyzer can act on the input files
    concurrently. single_threaded means that it should not be more than one active thread of the analyzer at a time.
    """
    multi_threaded = 1,
    single_threaded = 2


class Analyzer(ABC):
    """
    virtual class that represents an analizer program. It is assumed that the analyzer program acts on one or
    more WAGASCI runs. The output can be a series of plots, a new ROOT file or just to modify the input ROOT file or
    a combination of those.
    """
    depends = None

    def __init__(self,
                 analyzer_name,  # type: str
                 run_name,  # type: str
                 run_root_dir,  # type: str
                 output_dir,  # type: str
                 wagasci_libdir=None,  # type: Optional[str]
                 run_number=None,  # type: Optional[int]
                 default_args=None,  # type: Optional[Dict]
                 **kwargs):

        """
        :param analyzer_name: arbitrary name of the analyzer program
        :param run_name: name of the run to analyze
        :param run_root_dir: directory where the run files are located
        :param output_dir: directory where the output files are to be saved
        :param wagasci_libdir: directory where the WAGASCI library is
        :param run_number: run number
        :param default_args: default arguments of the analyzer program
        :param kwargs: variable arguments of the analyzer program
        """
        self.name = analyzer_name
        self.run_name = run_name
        self.run_number = run_number
        self.run_root_dir = run_root_dir
        self.output_dir = output_dir
        self._wagasci_libdir = wagasci_libdir
        if self._wagasci_libdir is None:
            try:
                env = environment.WagasciEnvironment()
                self._wagasci_libdir = env['WAGASCI_LIBDIR']
            except KeyError:
                raise KeyError("WAGASCI_LIBDIR variable not found in the shell environment")
        if default_args:
            self.args = default_args
        else:
            self.args = OrderedDict()
        self.set_init_arguments(**kwargs)

    @abc.abstractmethod
    def _set_runtime_arguments(self, input_files, dif_id):
        # type: (List[str], Optional[int]) -> None
        pass

    @abc.abstractmethod
    def spawn(self, chains):
        pass

    def get_topology(self, acq_config_xml):
        """
        Get the detector topology (DIF - CHIP - CHAN) from the XML configuration file
        :rtype: dict
        :param acq_config_xml: path to XML file containing the acquisition configuration
        :return: Detector topology dictionary
        """
        chain = analysis.WagasciAnalysis(self._wagasci_libdir)
        topology_string, pointer = chain.get_dif_topology(acq_config_xml)
        chain.free_topology(pointer)
        return json.loads(topology_string)

    def multiple_input_loop(self, input_files, chains):
        # type: (Dict[int, List[str]], Dict[int, Chain]) -> None
        """
        Analyze the input files concurrently  using the analyzer whose name is self.name and arguments are self.args
        :param input_files: input files
        :param chains: dictionary where the key is the DIF ID and the value is a Chain recordclass object.
        """
        # sanity checks
        print("Input files [DIF, input files path] : {}".format(input_files))
        if chains is None:
            raise ValueError("Chain recordclass objects not found")
        if not isinstance(chains, dict):
            raise TypeError("The chains dictionary must be initialized outside of the multiple_input_loop method")
        if len(chains) != 0 and len(chains) != len(input_files):
            raise ValueError("The number of chains ({}) must be the same as the number of "
                             "DIFs ({})".format(len(chains), len(input_files)))
        # DIF loop
        for dif_id, input_file in input_files.items():
            self._set_runtime_arguments(input_files=input_file, dif_id=utils.extract_dif_id(input_file))
            if dif_id not in chains:
                chain = Chain
                chain.link = analysis.WagasciAnalysis(self._wagasci_libdir)
                chain.thread = chain.link.spawn(self.name, **self.args)
                chains[dif_id] = chain
                time.sleep(1)
            else:
                chains[dif_id].thread = chains[dif_id].link.spawn(self.name, **self.args)
            print("Spawn thread with DIF {} : LINK ID {} : THREAD ID {}".format(dif_id, id(chains[dif_id].link),
                                                                                id(chains[dif_id].thread)))

    def single_input_loop(self, input_file, chains):
        # type: (str, Dict[int, Chain]) -> None
        """
        Analyze the input file using the analyzer whose name is self.name and arguments are self.args
        :param input_file: input file
        :param chains: dictionary where the key is the DIF ID and the value is a Chain recordclass object.
        """
        if chains and len(chains) > 1:
            raise IndexError("The method %s does not support more than one chain" %
                             inspect.currentframe().f_code.co_name)
        self._set_runtime_arguments(input_files=input_file, dif_id=utils.extract_dif_id(input_file))
        dummy_id = 0
        if not isinstance(chains, dict):
            raise TypeError("The chains dictionary must be initialized upstream")
        if dummy_id not in chains:
            chain = Chain
            chain.link = analysis.WagasciAnalysis(self._wagasci_libdir)
            chain.thread = chain.link.spawn(self.name, **self.args)
            chains[dummy_id] = chain
        else:
            chains[dummy_id].thread = chains[dummy_id].link.spawn(self.name, **self.args)
        print("Spawn thread with DIF {} : LINK ID {} : THREAD ID {}".format(dummy_id, id(chains[dummy_id].link),
                                                                            id(chains[dummy_id].thread)))

    def set_init_arguments(self, **kwargs):
        """
        Set the default value of some of the analyzer program arguments
        :param kwargs: default arguments
        """
        for key, value in kwargs.items():
            if key not in self.args:
                raise KeyError("Analyzer %s does not accept argument %s" % (self.name, key))
            self.args[key] = value


class Decoder(Analyzer):
    """
    Wrapper around the wgDecoder program
    """
    name = "decoder"
    depends = None
    input_type = AnalyzerInputType.single_run
    threading_type = AnalyzerThreadingType.multi_threaded

    _default_args = utils.get_arguments_ordered_dict(analysis.WagasciAnalysis.decoder)
    _default_args.update({'calibration_dir': "",
                          'overwrite_flag': False,
                          'compatibility_mode': False})

    def __init__(self, **kwargs):
        super(Decoder, self).__init__(analyzer_name=self.name, default_args=self._default_args, **kwargs)

        self.acq_config_xml = acqconfigxml_file_finder(self.run_root_dir, self.run_name)
        if not os.path.exists(self.acq_config_xml):
            raise OSError("Acquisition configuration XML file not found : %s" % self.acq_config_xml)
        self._topology = self.get_topology(self.acq_config_xml)
        utils.renametree(run_root_dir=self.run_root_dir, run_name=self.run_name, dif_topology=self._topology)

    def _set_runtime_arguments(self, input_files, dif_id):
        # type: (List[str], Optional[int]) -> None
        """
        Set arguments at execution time. By execution time I mean the moment the wgDecoder program is launched
        through the spawn method.
        :param dif_id: 
        :param input_files: input file
        :return: None
        """
        self.args["input_file"] = input_files
        self.args["output_dir"] = self.output_dir
        self.args["dif"] = dif_id
        self.args["n_chips"] = len(self._topology[str(dif_id)])

    def spawn(self, chains):
        # type: (Dict[int, Chain]) -> None
        """
        Spawn the wgDecoder threads contained in the chains dict
        :param chains: dictionary where the key is the DIF ID and the value is a Chain recordclass object.
        :return: None
        """
        if os.path.isfile(self.run_root_dir) and self.run_root_dir.endswith('.raw'):
            input_files = [utils.extract_dif_id(self.run_root_dir), [self.run_root_dir]]
        else:
            input_files = {utils.extract_dif_id(f): f for f in utils.find_files(self.run_root_dir, 'raw') if
                           utils.extract_dif_id(f) is not None}
        if not os.path.exists(self.output_dir):
            utils.mkdir_p(self.output_dir)
        for xml_file in utils.find_files(self.run_root_dir, 'xml'):
            dst = os.path.join(self.output_dir, os.path.basename(xml_file))
            if not os.path.exists(dst):
                shutil.copy2(src=xml_file, dst=dst)
        for log_file in utils.find_files(self.run_root_dir, 'log'):
            dst = os.path.join(self.output_dir, os.path.basename(log_file))
            if not os.path.exists(dst):
                shutil.copy2(src=log_file, dst=dst)
        self.multiple_input_loop(input_files, chains)


class MakeHist(Analyzer):
    """
    Wrapper around the wgMakeHist program
    """
    name = "make_hist"
    depends = Decoder.name
    input_type = AnalyzerInputType.single_run
    threading_type = AnalyzerThreadingType.single_threaded

    _default_args = utils.get_arguments_ordered_dict(analysis.WagasciAnalysis.make_hist)

    def __init__(self, **kwargs):
        super(MakeHist, self).__init__(analyzer_name=self.name, default_args=self._default_args, **kwargs)
        self.acq_config_xml = acqconfigxml_file_finder(self.run_root_dir, self.run_name)
        if not os.path.exists(self.acq_config_xml):
            raise OSError("Acquisition configuration XML file not found : %s" % self.acq_config_xml)
        self.args["config_file"] = self.acq_config_xml

    def _set_runtime_arguments(self, input_files, dif_id):
        # type: (List[str], Optional[int]) -> None
        """
        Set arguments at execution time. By execution time I mean the moment the wgMakeHist program is launched
        through the spawn method.
        :param dif_id: 
        :param input_files: input file
        :return: None
        """
        flags = bitarray('0' * 9, endian='big')
        flags[7] = True  # dark noise
        flags[4] = True  # charge hit HG
        flags[0] = False  # overwrite
        ul_flags = int(flags.to01(), 2)
        self.args["ul_flags"] = ul_flags
        self.args["input_file"] = input_files
        self.args["output_dir"] = os.path.join(self.output_dir, self.name)
        self.args["dif"] = dif_id

    def spawn(self, chains):
        # type: (Dict[int, Chain]) -> None
        """
        Spawn the wgMakeHist threads contained in the chains dict
        :param chains: dictionary where the key is the DIF ID and the value is a Chain recordclass object.
        :return: None
        """
        if os.path.isfile(self.run_root_dir) and self.run_root_dir.endswith('tree.root'):
            input_files = [self.run_root_dir]
        elif os.path.isdir(self.run_root_dir):
            input_files = filter(lambda f: f.endswith('tree.root'), utils.find_files(self.run_root_dir, 'root'))
        else:
            raise OSError("Input file or directory is invalid : {}".format(self.run_root_dir))
        if not os.path.exists(self.output_dir):
            utils.mkdir_p(self.output_dir)
        for inputfile in input_files:
            self.single_input_loop(inputfile, chains)


class TdcCalibration(Analyzer):
    """
    Wrapper around the wgTDC program
    """
    name = "tdc_calibration"
    depends = Decoder.name
    input_type = AnalyzerInputType.single_run
    threading_type = AnalyzerThreadingType.multi_threaded

    _default_args = utils.get_arguments_ordered_dict(analysis.WagasciAnalysis.tdc_calibration)

    def __init__(self, **kwargs):
        super(TdcCalibration, self).__init__(analyzer_name=self.name, default_args=self._default_args, **kwargs)
        self.acq_config_xml = acqconfigxml_file_finder(self.run_root_dir, self.run_name)
        if not os.path.exists(self.acq_config_xml):
            raise OSError("Acquisition configuration XML file not found : %s" % self.acq_config_xml)
        self.args["config_file"] = self.acq_config_xml

    def _set_runtime_arguments(self, input_files, dif_id):
        # type: (List[str], Optional[int]) -> None
        """
        Set arguments at execution time. By execution time I mean the moment the wgTDC program is launched
        through the spawn method.
        :param dif_id: 
        :param input_files: input file
        :return: None
        """
        self.args["input_file"] = input_files
        self.args["dif"] = dif_id

    def spawn(self, chains):
        # type: (Dict[int, Chain]) -> None
        """
        Spawn the wgTdcCalibration threads contained in the chains dict
        :param chains: dictionary where the key is the DIF ID and the value is a Chain recordclass object.
        :return: None
        """
        if os.path.isfile(self.run_root_dir) and self.run_root_dir.endswith('tree.root'):
            input_files = [self.run_root_dir]
        elif os.path.isdir(self.run_root_dir):
            input_files = filter(lambda f: f.endswith('tree.root'), utils.find_files(self.run_root_dir, 'root'))
            input_files = {utils.extract_dif_id(input_file): input_file for input_file in input_files}
        else:
            raise OSError("Input file or directory is invalid : {}".format(self.run_root_dir))
        self.multiple_input_loop(input_files, chains)


class SpillNumberFixer(Analyzer):
    """
    Wrapper around the wgSpillNumberFixer program
    """
    name = "spill_number_fixer"
    depends = Decoder.name
    input_type = AnalyzerInputType.single_run
    threading_type = AnalyzerThreadingType.single_threaded

    _default_args = utils.get_arguments_ordered_dict(analysis.WagasciAnalysis.spill_number_fixer)
    _default_args.update({'output_filename': "", 'passes': "", 'offset': 0, 'enable_graphics': False})

    def _set_runtime_arguments(self, input_files, dif_id):
        # type: (List[str], Optional[int]) -> None
        """
        Set arguments at execution time. By execution time I mean the moment the wgSpillNumberFixer program is launched
        through the spawn method.
        :param dif_id: 
        :return: None
        """
        self.args["output_filename"] = self.run_name
        self.args["passes"] = utils.spill_number_fixer_passes_calculator(self.run_number)
        if self.run_number >= 110:
            self.args["offset"] = 1

    def __init__(self, **kwargs):
        super(SpillNumberFixer, self).__init__(analyzer_name=self.name, default_args=self._default_args, **kwargs)
        self.args["input_dir"] = self.run_root_dir
        self.args["output_dir"] = self.output_dir

    def spawn(self, chains):
        """
        Spawn the wgSpillNumberFixer threads contained in the chains dict
        :param chains: dictionary where the key is the DIF ID and the value is a Chain recordclass object.
        :return: None
        """
        if not os.path.exists(self.output_dir):
            utils.mkdir_p(self.output_dir)
        self.single_input_loop(self.run_root_dir, chains)


class BeamSummaryData(Analyzer):
    """
    Wrapper around the wgBeamSummaryData program
    """
    name = "beam_summary_data"
    depends = SpillNumberFixer.name
    input_type = AnalyzerInputType.single_run
    threading_type = AnalyzerThreadingType.single_threaded

    _default_args = utils.get_arguments_ordered_dict(beam_summary_data.beam_summary_data)
    _default_args.update({'t2krun': 10, 'recursive': True})

    def __init__(self, **kwargs):
        super(BeamSummaryData, self).__init__(analyzer_name=self.name, default_args=self._default_args, **kwargs)

    def _set_runtime_arguments(self, input_files, dif_id):
        # type: (List[str], Optional[int]) -> None
        """
        Set arguments at execution time. By execution time I mean the moment the wgBeamSummaryData program is launched
        through the spawn method.
        :param dif_id: 
        :return: None
        """
        self.args["input_path"] = input_files
        if os.path.isdir(input_files):
            self.args["recursive"] = True
        else:
            self.args["recursive"] = False

    def spawn(self, chains):
        # type: (Dict[int, Chain]) -> None
        """
        Spawn the wgBeamSummaryData threads contained in the chains dict
        :param chains: dictionary where the key is the DIF ID and the value is a Chain recordclass object.
        :return: None
        """
        self.single_input_loop(self.run_root_dir, chains)


class ApplyDetectorFlags(Analyzer):
    """
    Wrapper around the wgApplyDetectorFlags program
    """
    name = "apply_detector_flags"
    depends = SpillNumberFixer.name
    input_type = AnalyzerInputType.single_run
    threading_type = AnalyzerThreadingType.single_threaded

    _default_args = utils.get_arguments_ordered_dict(apply_detector_flags.apply_detector_flags)

    def __init__(self, **kwargs):
        super(ApplyDetectorFlags, self).__init__(analyzer_name=self.name, default_args=self._default_args, **kwargs)

    def _set_runtime_arguments(self, input_files, dif_id):
        # type: (List[str], Optional[int]) -> None
        """
        Set arguments at execution time. By execution time I mean the moment the program is launched
        through the spawn method.
        :param dif_id: 
        :return: None
        """
        self.args["input_path"] = input_files

    def spawn(self, chains):
        # type: (Dict[int, Chain]) -> None
        """
        Spawn the threads contained in the chains dict
        :param chains: dictionary where the key is the DIF ID and the value is a Chain recordclass object.
        :return: None
        """
        self.single_input_loop(self.run_root_dir, chains)


class BcidDistribution(Analyzer):
    """
    Wrapper around the bcid_distribution function
    """
    name = "bcid_distribution"
    depends = MakeHist.name
    input_type = AnalyzerInputType.single_run
    threading_type = AnalyzerThreadingType.single_threaded

    _default_args = utils.get_arguments_ordered_dict(analysis.WagasciAnalysis.bcid_distribution)
    _default_args.update({'chip_by_chip': True})

    def __init__(self, **kwargs):
        super(BcidDistribution, self).__init__(analyzer_name=self.name, default_args=self._default_args, **kwargs)

    def _set_runtime_arguments(self, input_files, dif_id):
        # type: (List[str], Optional[int]) -> None
        """
        Set arguments at execution time. By execution time I mean the moment the bcid distribution function  is
        executed through the spawn method.
        :param dif_id: 
        :return: None
        """
        self.args["input_file"] = input_files
        self.args["output_img_dir"] = os.path.join(self.output_dir, self.name)

    def spawn(self, chains):
        # type: (Dict[int, Chain]) -> None
        """
        Spawn the wgBCID threads contained in the chains dict
        :param chains: dictionary where the key is the DIF ID and the value is a Chain recordclass object.
        :return: None
        """
        if os.path.isfile(self.run_root_dir) and self.run_root_dir.endswith('hist.root'):
            input_files = [self.run_root_dir]
        else:
            make_hist_dir = os.path.join(self.run_root_dir, MakeHist.name)
            input_files = utils.find_files(make_hist_dir, 'root')

        self.args["topology_source"] = acqconfigxml_file_finder(self.run_root_dir, os.path.basename(self.run_root_dir))

        if not os.path.exists(self.output_dir):
            utils.mkdir_p(self.output_dir)
        for input_file in input_files:
            self._set_runtime_arguments(input_files=input_file, dif_id=None)
            link = analysis.WagasciAnalysis(self._wagasci_libdir)
            link.bcid_distribution(**self.args)


class AdcDistribution(Analyzer):
    """
    Wrapper around the adc_distribution function
    """
    name = "adc_distribution"
    depends = MakeHist.name
    input_type = AnalyzerInputType.single_run
    threading_type = AnalyzerThreadingType.single_threaded

    _default_args = utils.get_arguments_ordered_dict(analysis.WagasciAnalysis.adc_distribution)
    _default_args.update({'chip_by_chip': True})

    def __init__(self, **kwargs):
        super(AdcDistribution, self).__init__(analyzer_name=self.name, default_args=self._default_args, **kwargs)

    def _set_runtime_arguments(self, input_files, dif_id):
        # type: (List[str], Optional[int]) -> None
        """
        Set arguments at execution time. By execution time I mean the moment the adc distribution function  is
        executed through the spawn method.
        :param dif_id: 
        :return: None
        """
        self.args["input_file"] = input_files
        self.args["output_img_dir"] = os.path.join(self.output_dir, self.name)

    def spawn(self, chains):
        # type: (Dict[int, Chain]) -> None
        """
        Spawn the wgADC threads contained in the chains dict
        :param chains: dictionary where the key is the DIF ID and the value is a Chain recordclass object.
        :return: None
        """
        if os.path.isfile(self.run_root_dir) and self.run_root_dir.endswith('hist.root'):
            input_files = [self.run_root_dir]
        else:
            make_hist_dir = os.path.join(self.run_root_dir, MakeHist.name)
            input_files = utils.find_files(make_hist_dir, 'root')

        self.args["topology_source"] = acqconfigxml_file_finder(self.run_root_dir, os.path.basename(self.run_root_dir))

        if not os.path.exists(self.output_dir):
            utils.mkdir_p(self.output_dir)
        for input_file in input_files:
            self._set_runtime_arguments(input_files=input_file, dif_id=None)
            link = analysis.WagasciAnalysis(self._wagasci_libdir)
            link.adc_distribution(**self.args)


class Temperature(Analyzer):
    """
    Wrapper around the wgTemperature program
    """
    name = "temperature"
    depends = BeamSummaryData.name
    input_type = AnalyzerInputType.single_run
    threading_type = AnalyzerThreadingType.multi_threaded

    _default_args = utils.get_arguments_ordered_dict(analysis.WagasciAnalysis.temperature)
    _default_args.update({'sqlite_database': "/hsm/nu/wagasci/data/temphum/mh_temperature_sensors_t2krun10.sqlite3"})

    def __init__(self, **kwargs):
        super(Temperature, self).__init__(analyzer_name=self.name, default_args=self._default_args, **kwargs)

    def _set_runtime_arguments(self, input_files, dif_id):
        # type: (List[str], Optional[int]) -> None
        """
        Set arguments at execution time. By execution time I mean the moment the wgTemperature program is launched
        through the spawn method.
        :param dif_id: 
        :param input_files: input file
        :return: None
        """
        self.args["input_file"] = input_files

    def spawn(self, chains):
        # type: (Dict[int, Chain]) -> None
        """
        Spawn the wgTemperature threads contained in the chains dict
        :param chains: dictionary where the key is the DIF ID and the value is a Chain recordclass object.
        :return: None
        """
        if os.path.isfile(self.run_root_dir) and self.run_root_dir.endswith('.root'):
            input_files = [utils.extract_dif_id(self.run_root_dir), [self.run_root_dir]]
        else:
            input_files = {utils.extract_dif_id(f): f for f in utils.find_files(self.run_root_dir, 'root') if
                           utils.extract_dif_id(f) is not None}
        self.multiple_input_loop(input_files, chains)


class AdcCalibration(Analyzer):
    """
    Wrapper around the wgAdcCalib program (all passes)
    """
    name = "adc_calibration"
    depends = Temperature.name
    input_type = AnalyzerInputType.multiple_runs
    threading_type = AnalyzerThreadingType.multi_threaded

    _default_args = utils.get_arguments_ordered_dict(analysis.WagasciAnalysis.adc_calibration)
    _default_args.update({'passes': -1, 'enable_plotting': False, 'fixed_wallmrd_gain': False,
                          'fixed_wallmrd_input_dac': 121, 'silent_mode': True})

    def __init__(self, **kwargs):
        super(AdcCalibration, self).__init__(analyzer_name=self.name, default_args=self._default_args, **kwargs)
        self.history_filename = self.run_name
        self.args["history_dir"] = self.output_dir
        if self.run_number is not None:
            self.history_filename = "{}_{}".format(self.history_filename, self.run_number)

    def _set_runtime_arguments(self, input_files, dif_id):
        # type: (List[str], Optional[int]) -> None
        """
        Set arguments at execution time. By execution time I mean the moment the wgAdcCalib program is launched
        through the spawn method.
        :param dif_id: 
        :param input_files: input file
        :return: None
        """
        self.args["tree_files"] = input_files
        self.args["history_file"] = "{}_ecal_dif_{}.root".format(self.history_filename, dif_id)
        self.args["dif_id"] = dif_id

    def spawn(self, chains):
        # type: (Dict[int, Chain]) -> None
        """
        Spawn the wgAdcCalib threads contained in the chains dict
        :param chains: dictionary where the key is the DIF ID and the value is a Chain recordclass object.
        :return: None
        """
        if not isinstance(self.run_root_dir, list):
            self.run_root_dir = [self.run_root_dir]
        input_files = {}
        self.args["topology_source"] = acqconfigxml_file_finder(self.run_root_dir[0],
                                                                os.path.basename(self.run_root_dir[0]))
        topology = self.get_topology(self.args["topology_source"])
        for dif_id in [int(dif_id) for dif_id in topology]:
            input_files[dif_id] = []
            for run_dir in sorted(self.run_root_dir):
                for root_file in utils.find_files(path=run_dir, extension='root', recursive=False):
                    if dif_id == utils.extract_dif_id(root_file):
                        input_files[dif_id].append(root_file)
                        break
        self.multiple_input_loop(input_files, chains)


class SanityCheck(Analyzer):
    """
    Wrapper around the sanity_check function
    """
    name = "sanity_check"
    depends = AdcCalibration.name
    input_type = AnalyzerInputType.single_run
    threading_type = AnalyzerThreadingType.single_threaded

    _default_args = utils.get_arguments_ordered_dict(sanity_check.sanity_check)

    def __init__(self, **kwargs):
        super(SanityCheck, self).__init__(analyzer_name=self.name, default_args=self._default_args, **kwargs)

    def _set_runtime_arguments(self, input_files, dif_id):
        # type: (List[str], Optional[int]) -> None
        """
        Set arguments at execution time. By execution time I mean the moment the program is launched
        through the spawn method.
        :param dif_id: 
        :return: None
        """
        self.args["input_path"] = input_files

    def spawn(self, chains):
        # type: (Dict[int, Chain]) -> None
        """
        Spawn the threads contained in the chains dict
        :param chains: dictionary where the key is the DIF ID and the value is a Chain recordclass object.
        :return: None
        """
        self.single_input_loop(self.run_root_dir, chains)


class AnalyzerFactory(ABC):
    """
    Abstract factory design patter to produce Analyzer objects.
    """

    def __init__(self, **kwargs):
        """
        :param kwargs: arguments to pass to the analyzer object constructor
        """
        self._kwargs = kwargs

    @abc.abstractmethod
    def get_analyzer(self, run_root_dir, output_dir, **kwargs):
        # type: (str, str, **Any) -> None
        """
        Build analyzer (abstract method)
        :param run_root_dir: directory where the run files are stored
        :param output_dir: directory where to save the output files
        :param kwargs: additional arguments for the analyzer
        :return: None
        """
        self._kwargs.update(kwargs)


class DecoderFactory(AnalyzerFactory):
    depends = Decoder.depends
    name = Decoder.name
    input_type = Decoder.input_type
    threading_type = Decoder.threading_type

    def get_analyzer(self, run_root_dir, output_dir, **kwargs):
        # type: (str, str, **Any) -> Decoder
        """
        Build wgDecoder analyzer
        :param run_root_dir: directory where the run files are stored
        :param output_dir: directory where to save the output files
        :param kwargs: additional arguments for the analyzer
        :return: Analyzer object
        """
        super(DecoderFactory, self).get_analyzer(run_root_dir=run_root_dir, output_dir=output_dir, **kwargs)
        return Decoder(run_root_dir=run_root_dir, output_dir=output_dir, **self._kwargs)


class MakeHistFactory(AnalyzerFactory):
    depends = MakeHist.depends
    name = MakeHist.name
    input_type = MakeHist.input_type
    threading_type = MakeHist.threading_type

    def get_analyzer(self, run_root_dir, output_dir, **kwargs):
        # type: (str, str, **Any) -> MakeHist
        """
        Build wgMakeHist analyzer
        :param run_root_dir: directory where the run files are stored
        :param output_dir: directory where to save the output files
        :param kwargs: additional arguments for the analyzer
        :return: MakeHist object
        """
        super(MakeHistFactory, self).get_analyzer(run_root_dir=run_root_dir, output_dir=output_dir, **kwargs)
        return MakeHist(run_root_dir=run_root_dir, output_dir=output_dir, **self._kwargs)


class TdcCalibrationFactory(AnalyzerFactory):
    depends = TdcCalibration.depends
    name = TdcCalibration.name
    input_type = TdcCalibration.input_type
    threading_type = TdcCalibration.threading_type

    def get_analyzer(self, run_root_dir, output_dir, **kwargs):
        # type: (str, str, **Any) -> TdcCalibration
        """
        Build wgTdcCalibration analyzer
        :param run_root_dir: directory where the run files are stored
        :param output_dir: directory where to save the output files
        :param kwargs: additional arguments for the analyzer
        :return: TdcCalibration object
        """
        super(TdcCalibrationFactory, self).get_analyzer(run_root_dir=run_root_dir, output_dir=output_dir, **kwargs)
        return TdcCalibration(run_root_dir=run_root_dir, output_dir=output_dir, **self._kwargs)


class SpillNumberFixerFactory(AnalyzerFactory):
    depends = SpillNumberFixer.depends
    name = SpillNumberFixer.name
    input_type = SpillNumberFixer.input_type
    threading_type = SpillNumberFixer.threading_type

    def get_analyzer(self, run_root_dir, output_dir, **kwargs):
        # type: (str, str, **Any) -> SpillNumberFixer
        """
        Build wgSpillNumberFixer analyzer
        :param run_root_dir: directory where the run files are stored
        :param output_dir: directory where to save the output files
        :param kwargs: additional arguments for the analyzer
        :return: Analyzer object
        """
        super(SpillNumberFixerFactory, self).get_analyzer(run_root_dir=run_root_dir, output_dir=output_dir, **kwargs)
        return SpillNumberFixer(run_root_dir=run_root_dir, output_dir=output_dir, **self._kwargs)


class BeamSummaryDataFactory(AnalyzerFactory):
    depends = BeamSummaryData.depends
    name = BeamSummaryData.name
    input_type = BeamSummaryData.input_type
    threading_type = BeamSummaryData.threading_type

    def get_analyzer(self, run_root_dir, output_dir, **kwargs):
        # type: (str, str, **Any) -> BeamSummaryData
        """
        Build beam_summary_data analyzer
        :param run_root_dir: directory where the run files are stored
        :param output_dir: directory where to save the output files
        :param kwargs: additional arguments for the analyzer
        :return: Analyzer object
        """
        super(BeamSummaryDataFactory, self).get_analyzer(run_root_dir=run_root_dir, output_dir=output_dir, **kwargs)
        return BeamSummaryData(run_root_dir=run_root_dir, output_dir=output_dir, **self._kwargs)


class ApplyDetectorFlagsFactory(AnalyzerFactory):
    depends = ApplyDetectorFlags.depends
    name = ApplyDetectorFlags.name
    input_type = ApplyDetectorFlags.input_type
    threading_type = ApplyDetectorFlags.threading_type

    def get_analyzer(self, run_root_dir, output_dir, **kwargs):
        # type: (str, str, **Any) -> ApplyDetectorFlags
        """
        Build apply_detector_flags analyzer
        :param run_root_dir: directory where the run files are stored
        :param output_dir: directory where to save the output files
        :param kwargs: additional arguments for the analyzer
        :return: Analyzer object
        """
        super(ApplyDetectorFlagsFactory, self).get_analyzer(run_root_dir=run_root_dir, output_dir=output_dir, **kwargs)
        return ApplyDetectorFlags(run_root_dir=run_root_dir, output_dir=output_dir, **self._kwargs)


class SanityCheckFactory(AnalyzerFactory):
    depends = SanityCheck.depends
    name = SanityCheck.name
    input_type = SanityCheck.input_type
    threading_type = SanityCheck.threading_type

    def get_analyzer(self, run_root_dir, output_dir, **kwargs):
        # type: (str, str, **Any) -> SanityCheck
        """
        Build sanity_check analyzer
        :param run_root_dir: directory where the run files are stored
        :param output_dir: directory where to save the output files
        :param kwargs: additional arguments for the analyzer
        :return: Analyzer object
        """
        super(SanityCheckFactory, self).get_analyzer(run_root_dir=run_root_dir, output_dir=output_dir, **kwargs)
        return SanityCheck(run_root_dir=run_root_dir, output_dir=output_dir, **self._kwargs)


class BcidDistributionFactory(AnalyzerFactory):
    depends = BcidDistribution.depends
    name = BcidDistribution.name
    input_type = BcidDistribution.input_type
    threading_type = BcidDistribution.threading_type

    def get_analyzer(self, run_root_dir, output_dir, **kwargs):
        # type: (str, str, **Any) -> BcidDistribution
        """
        Build bcid_distribution analyzer
        :param run_root_dir: directory where the run files are stored
        :param output_dir: directory where to save the output files
        :param kwargs: additional arguments for the analyzer
        :return: Analyzer object
        """
        super(BcidDistributionFactory, self).get_analyzer(run_root_dir=run_root_dir, output_dir=output_dir, **kwargs)
        return BcidDistribution(run_root_dir=run_root_dir, output_dir=output_dir, **self._kwargs)


class AdcDistributionFactory(AnalyzerFactory):
    depends = AdcDistribution.depends
    name = AdcDistribution.name
    input_type = AdcDistribution.input_type
    threading_type = AdcDistribution.threading_type

    def get_analyzer(self, run_root_dir, output_dir, **kwargs):
        # type: (str, str, **Any) -> AdcDistribution
        """
        Build adc_distribution analyzer
        :param run_root_dir: directory where the run files are stored
        :param output_dir: directory where to save the output files
        :param kwargs: additional arguments for the analyzer
        :return: Analyzer object
        """
        super(AdcDistributionFactory, self).get_analyzer(run_root_dir=run_root_dir, output_dir=output_dir, **kwargs)
        return AdcDistribution(run_root_dir=run_root_dir, output_dir=output_dir, **self._kwargs)


class TemperatureFactory(AnalyzerFactory):
    depends = Temperature.depends
    name = Temperature.name
    input_type = Temperature.input_type
    threading_type = Temperature.threading_type

    def get_analyzer(self, run_root_dir, output_dir, **kwargs):
        # type: (str, str, **Any) -> Temperature
        """
        Build wgTemperature analyzer
        :param run_root_dir: directory where the run files are stored
        :param output_dir: directory where to save the output files
        :param kwargs: additional arguments for the analyzer
        :return: Analyzer object
        """
        super(TemperatureFactory, self).get_analyzer(run_root_dir=run_root_dir, output_dir=output_dir, **kwargs)
        return Temperature(run_root_dir=run_root_dir, output_dir=output_dir, **self._kwargs)


class AdcCalibrationFactory(AnalyzerFactory):
    depends = AdcCalibration.depends
    name = AdcCalibration.name
    input_type = AdcCalibration.input_type
    threading_type = AdcCalibration.threading_type

    def get_analyzer(self, run_root_dir, output_dir, **kwargs):
        # type: (str, str, **Any) -> AdcCalibration
        """
        Build wgAdcCalib analyzer
        :param run_root_dir: directory where the run files are stored
        :param output_dir: directory where to save the output files
        :param kwargs: additional arguments for the analyzer
        :return: Analyzer object
        """
        super(AdcCalibrationFactory, self).get_analyzer(run_root_dir=run_root_dir, output_dir=output_dir, **kwargs)
        return AdcCalibration(run_root_dir=run_root_dir, output_dir=output_dir, **self._kwargs)


class AnalyzerFactoryProducer:
    """
    Abstract factory design patter to produce Analyzer objects.
    """
    _ReturnType = Union[DecoderFactory, MakeHistFactory, TdcCalibrationFactory, SpillNumberFixerFactory,
                        BeamSummaryDataFactory, ApplyDetectorFlagsFactory, TemperatureFactory, AdcCalibrationFactory,
                        BcidDistributionFactory, AdcDistributionFactory, SanityCheckFactory]

    def __init__(self, wagasci_libdir=None):
        # type: (str) -> None
        """
        :param wagasci_libdir: WAGASCI library directory
        """
        self._wagasci_libdir = wagasci_libdir
        pass

    def get_factory(self, type_of_factory, **kwargs):
        # type: (str, **Any) -> _ReturnType
        """
        Return the analyzer factory of the desired type
        :param type_of_factory: name of the factory
        :param kwargs: arguments to pass to the factory constructor
        :return: AnalyzerFactory object
        """
        if type_of_factory == Decoder.name:
            return DecoderFactory(wagasci_libdir=self._wagasci_libdir, **kwargs)
        elif type_of_factory == MakeHist.name:
            return MakeHistFactory(wagasci_libdir=self._wagasci_libdir, **kwargs)
        elif type_of_factory == TdcCalibration.name:
            return TdcCalibrationFactory(wagasci_libdir=self._wagasci_libdir, **kwargs)
        elif type_of_factory == SpillNumberFixer.name:
            return SpillNumberFixerFactory(wagasci_libdir=self._wagasci_libdir, **kwargs)
        elif type_of_factory == BeamSummaryData.name:
            return BeamSummaryDataFactory(wagasci_libdir=self._wagasci_libdir, **kwargs)
        elif type_of_factory == ApplyDetectorFlags.name:
            return ApplyDetectorFlagsFactory(wagasci_libdir=self._wagasci_libdir, **kwargs)
        elif type_of_factory == Temperature.name:
            return TemperatureFactory(wagasci_libdir=self._wagasci_libdir, **kwargs)
        elif type_of_factory == AdcCalibration.name:
            return AdcCalibrationFactory(wagasci_libdir=self._wagasci_libdir, **kwargs)
        elif type_of_factory == BcidDistribution.name:
            return BcidDistributionFactory(wagasci_libdir=self._wagasci_libdir, **kwargs)
        elif type_of_factory == AdcDistribution.name:
            return AdcDistributionFactory(wagasci_libdir=self._wagasci_libdir, **kwargs)
        elif type_of_factory == SanityCheck.name:
            return SanityCheckFactory(wagasci_libdir=self._wagasci_libdir, **kwargs)
        raise NotImplementedError("Factory %s not implemented or not recognized" % type_of_factory)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2019 Pintaudi Giorgio, Eguchi Aoi

# pylint: disable-msg=too-many-arguments
# pylint: disable-msg=too-many-locals
# pylint: disable-msg=too-many-branches

"""
WAGASCI Analysis Python module

This module is just a wrapping around some of the WAGASCI Analysis C++ libraries.
The libraries are wrapped using Ctypes.

You can call the analysis methods directly as blocking functions or
you can spawn a thread and concatenate how many analysis methods as
you want. When chained, each method is not executed until the previous
one has returned. If there is an error or an exception in any link of
the chain all the successive methods are not executed. If you want,
you can wait for the last method in the chain to return and check its
return value by explicitly using the join() method.

Usage examples:

    # Call analysis methods as blocking functions

    process = WagasciAnalysis("lib_dir")
    process.function1("argument1")
    process.function2("argument2")
    del process # optional

    # Execute function1, function2 and function3 in sequence without blocking
    # the main process and then wait for function3 to return

    process = WagasciAnalysis("lib_dir")
    process.spawn("function1","argument1")
    process.spawn("function2","argument2")
    thread = process.spawn("function3","argument3")
    del process # optional
    # Do other stuff
    result = thread.join()
    if result not 0 :
        print "Error number : %d" %(result)

"""

import ctypes as ct
import json
import os
import sys
import threading
from typing import Union, List, Tuple

from bitarray import bitarray

import wagascianpy.analysis.apply_detector_flags
import wagascianpy.analysis.beam_summary_data
import wagascianpy.analysis.sanity_check
import wagascianpy.utils.utils
from wagascianpy.utils.environment import WagasciEnvironment

OPTIMIZE_THRESHOLD_MODE = 0
OPTIMIZE_INPUTDAC_MODE = 1
SPIROC2D_NCHANNELS = 36
WAGASCI_NCHIPS = 20
WALLMRD_NCHIPS = 3


def _wrap_function(lib, funcname, restype, argtypes):
    """Simplify wrapping ctypes functions"""
    func = lib.__getattr__(funcname)
    func.restype = restype
    func.argtypes = argtypes
    return func


# noinspection PyUnresolvedReferences
class ThreadWithReturnValue(threading.Thread):
    """Thread subclass where the join() method returns the return value of the
    called function.

    This class inherits from the threading. Thread class and the only difference
    is that its join() method returns the return value of the called
    function. The return value is always zero, a positive integer or minus
    one. Zero means the the called function returned successfully. A positive
    integer means that there was an error in the called function. Minus one
    means that there was an error in one of the chained functions.

    """

    # noinspection PyArgumentList
    def __init__(self, group=None, target=None, name=None, args=(), kwargs=None, daemon=None):
        if sys.version_info.major >= 3:
            threading.Thread.__init__(self, group, target, name, args, kwargs, daemon=daemon)
        else:
            threading.Thread.__init__(self, group, target, name, args, kwargs, verbose=False)
        self._return = None

    def run(self):
        if sys.version_info.major >= 3:
            if self._target is not None:
                self._return = self._target(*self._args, **self._kwargs)
        else:
            if self._Thread__target is not None:
                self._return = self._Thread__target(*self._Thread__args, **self._Thread__kwargs)

    def join(self, timeout=None):
        threading.Thread.join(self, timeout)
        return self._return


class WagasciAnalysis(object):
    """Class to manage WAGASCI Analysis processes. Refer to the WAGASCI Analysis
    software documentation for further info.

    """

    def __init__(self, lib_dir, thread=None):
        if lib_dir is not None and os.path.isdir(lib_dir):
            self.m_lib_dir = lib_dir.rstrip('/')
        else:
            raise ValueError("WAGASCI library directory could not be found : %s" % lib_dir)
        self._previous_thread = thread
        self._result = 0
        self._previous_process = ""
        self.lock = threading.Lock()

    # The _dispatcher method calls the method called "function". If the "thread"
    # argument is different from null, it waits until that thread is over before
    # calling the function.
    def _dispatcher(self, thread, function, *argv, **kwargs):
        # Acquire lock to access the resources _result and _previous_process
        self.lock.acquire()
        try:
            # Wait for the previous thread to finish
            if thread is not None:
                self._result = thread.join()
            # if the _result attribute has been already set to -1 it means that
            # a previous thread returned an error. We can therefore avoid
            # calling all the successive functions
            if self._result == -1:
                raise RuntimeWarning
            elif self._result != 0:
                if self._result is not None:
                    raise ValueError("[%s] Error number : %d"
                                     % (self._previous_process, self._result))
                raise ValueError("[%s] Unspecified error"
                                 % self._previous_process)
            # Call the method named "function" and pass all the arguments "argv"
            # to it
            else:
                print("%s : start function %s" % (id(self), function))
                self._result = getattr(self, function)(*argv, **kwargs)
                print("%s : end function %s" % (id(self), function))
        except ValueError as error:
            print(str(error))
            self._result = -1
        finally:
            self._previous_process = function
            self.lock.release()
        return self._result

    def spawn(self, function, *argv, **kwargs):
        """Spawn a new thread that calls "function" whose arguments are "argv".
        Multiple threads can be chained together by calling the spawn function
        multiple times for the same object. This is a non-blocking function.

        Arguments:
        function = name of method to call
        argv = arguments to pass to *function*

        Return:
        object of the class ThreadWithReturnValue

        """
        argv = list(argv)
        argv.insert(0, self._previous_thread)
        argv.insert(1, function)
        self._previous_thread = ThreadWithReturnValue(target=self._dispatcher, args=tuple(argv), kwargs=kwargs)
        self._previous_thread.start()
        return self._previous_thread

    @wagascianpy.utils.utils.utf8_decorator
    def decoder(self, input_file, calibration_dir, output_dir, overwrite_flag, compatibility_mode, dif, n_chips):
        # type: (str, str, str, bool, bool, int, int) -> ThreadWithReturnValue
        """Wrapping for the wgDecoder process"""
        lib_decoder = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgDecoder.so'))
        wg_decoder = _wrap_function(lib_decoder, 'wgDecoder', ct.c_int,
                                    [ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_bool, ct.c_bool, ct.c_int, ct.c_int])
        return wg_decoder(input_file, calibration_dir, output_dir, overwrite_flag, compatibility_mode, dif, n_chips)

    @wagascianpy.utils.utils.utf8_decorator
    def change_config(self, input_file, output_file, overwrite_flag, edit_flag, value, mode,
                      channel=SPIROC2D_NCHANNELS):
        # type: (str, str, bool, bool, int, int, int) -> ThreadWithReturnValue
        """Wrapping for the wgChangeConfig process"""
        flags = bitarray('0' * 2, endian='big')
        flags[1] = bool(edit_flag)
        flags[0] = bool(overwrite_flag)
        flags_ulong = int(flags.to01(), 2)
        lib_change_config = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgChangeConfig.so'))
        wg_change_config = _wrap_function(lib_change_config, 'wgChangeConfig', ct.c_int,
                                          [ct.c_char_p, ct.c_char_p, ct.c_ulong, ct.c_int, ct.c_uint, ct.c_uint])
        return wg_change_config(input_file, output_file, flags_ulong, int(value), int(mode), int(channel))

    @wagascianpy.utils.utils.utf8_decorator
    def optimize(self, threshold_card, gain_card, topology_source, bitstream_config_dir, mode=OPTIMIZE_INPUTDAC_MODE,
                 photo_electrons=2, input_dac=255, adj_th=0):
        # type: (str, str, str, str, int, int, int, int) -> ThreadWithReturnValue
        """Wrapping for the wgOptimize process"""
        lib_optimize = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgOptimize.so'))
        wg_optimize = _wrap_function(lib_optimize, 'wgOptimize', ct.c_int,
                                     [ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_uint, ct.c_uint,
                                      ct.c_uint, ct.c_uint])
        return wg_optimize(threshold_card, gain_card, topology_source, bitstream_config_dir, mode, photo_electrons,
                           input_dac, adj_th)

    @wagascianpy.utils.utils.utf8_decorator
    def make_hist(self, input_file, config_file, output_dir, ul_flags, dif):
        # type: (str, str, str, int, int) -> ThreadWithReturnValue
        """Wrapping for the wgMakeHist process"""
        lib_make_hist = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgMakeHist.so'))
        wg_make_hist = _wrap_function(lib_make_hist, 'wgMakeHist', ct.c_int,
                                      [ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_ulong, ct.c_uint])
        return wg_make_hist(input_file, config_file, output_dir, ul_flags, dif)

    @wagascianpy.utils.utils.utf8_decorator
    def ana_hist(self, input_file, config_file, output_dir, output_img_dir, ul_flags, dif):
        # type: (str, str, str, str, int, int) -> ThreadWithReturnValue
        """Wrapping for the wgAnaHist process"""
        lib_ana_hist = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgAnaHist.so'))
        wg_ana_hist = _wrap_function(lib_ana_hist, 'wgAnaHist', ct.c_int,
                                     [ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_ulong, ct.c_uint])
        return wg_ana_hist(input_file, config_file, output_dir, output_img_dir, ul_flags, dif)

    @wagascianpy.utils.utils.utf8_decorator
    def ana_hist_summary(self, input_dir, output_xml_dir, output_img_dir, ul_flags):
        # type: (str, str, str, int) -> ThreadWithReturnValue
        """Wrapping for the wgAnaHistSummary process"""
        lib_ana_hist_summary = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgAnaHistSummary.so'))
        wg_ana_hist_summary = _wrap_function(lib_ana_hist_summary, 'wgAnaHistSummary', ct.c_int,
                                             [ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_ulong])
        return wg_ana_hist_summary(input_dir, output_xml_dir, output_img_dir, ul_flags)

    @wagascianpy.utils.utils.utf8_decorator
    def pedestal_calib(self, input_dir, output_xml_dir, output_img_dir):
        # type: (str, str, str) -> ThreadWithReturnValue
        """Wrapping for the wgPedestalCalib process"""
        lib_pedestal_calib = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgPedestalCalib.so'))
        wg_pedestal_calib = _wrap_function(lib_pedestal_calib, 'wgPedestalCalib',
                                           ct.c_int,
                                           [ct.c_char_p, ct.c_char_p, ct.c_char_p])
        return wg_pedestal_calib(input_dir, output_xml_dir, output_img_dir)

    @wagascianpy.utils.utils.utf8_decorator
    def gain_calib(self, input_dir, output_xml_dir, output_img_dir, print_flag):
        # type: (str, str, str, bool) -> ThreadWithReturnValue
        """Wrapping for the wgGainCalib process"""
        lib_gain_calib = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgGainCalib.so'))
        wg_gain_calib = _wrap_function(lib_gain_calib, 'wgGainCalib', ct.c_int,
                                       [ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_bool])
        return wg_gain_calib(input_dir, output_xml_dir, output_img_dir, print_flag)

    @wagascianpy.utils.utils.utf8_decorator
    def gain_check(self, input_dir, xml_config_file, output_img_dir, only_wallmrd, only_wagasci):
        # type: (str, str, str, bool, bool) -> ThreadWithReturnValue
        """Wrapping for the wgGainCheck process"""
        lib_gain_check = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgGainCheck.so'))
        wg_gain_check = _wrap_function(lib_gain_check, 'wgGainCheck', ct.c_int,
                                       [ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_bool, ct.c_bool])
        return wg_gain_check(input_dir, xml_config_file, output_img_dir, only_wallmrd, only_wagasci)

    @wagascianpy.utils.utils.utf8_decorator
    def gain_tune(self, input_dir, xml_config_file, output_img_dir, only_wallmrd, only_wagasci):
        # type: (str, str, str, bool, bool) -> ThreadWithReturnValue
        """Wrapping for the wgGainTune process"""
        lib_gain_tune = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgGainTune.so'))
        wg_gain_tune = _wrap_function(lib_gain_tune, 'wgGainTune', ct.c_int,
                                      [ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_bool, ct.c_bool])
        return wg_gain_tune(input_dir, xml_config_file, output_img_dir, only_wallmrd, only_wagasci)

    @wagascianpy.utils.utils.utf8_decorator
    def scurve(self, input_dir, output_xml_dir, output_img_dir, paranoid_mode):
        # type: (str, str, str, bool) -> ThreadWithReturnValue
        """Wrapping for the wgScurve process"""
        lib_scurve = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgScurve.so'))
        wg_scurve = _wrap_function(lib_scurve, 'wgScurve', ct.c_int,
                                   [ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_bool])
        return wg_scurve(input_dir, output_xml_dir, output_img_dir, paranoid_mode)

    @wagascianpy.utils.utils.utf8_decorator
    def spill_number_fixer(self, input_dir, output_dir, output_filename, passes, offset, enable_graphics):
        # type: (str, str, str, str, int, bool) -> ThreadWithReturnValue
        """Wrapping for the wgSpillNumberFixer process"""
        lib_snf = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgSpillNumberFixer.so'))
        wg_snf = _wrap_function(lib_snf, 'wgSpillNumberFixer', ct.c_int,
                                [ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_int, ct.c_bool])
        return wg_snf(input_dir, output_dir, output_filename, passes, offset, enable_graphics)

    @wagascianpy.utils.utils.utf8_decorator
    def temperature(self, sqlite_database, input_file):
        # type: (str, str) -> ThreadWithReturnValue
        """Wrapping for the wgTemperature process"""
        lib_temperature = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgTemperature.so'))
        wg_temperature = _wrap_function(lib_temperature, 'wgTemperature', ct.c_int, [ct.c_char_p, ct.c_char_p])
        return wg_temperature(sqlite_database, input_file)

    def adc_calibration(self, tree_files, topology_source, history_dir, history_file, dif_id, passes,
                        enable_plotting, fixed_wallmrd_gain, fixed_wallmrd_input_dac, silent_mode):
        # type: (Union[List, str], str, str, str, int, int, bool, bool, int, bool) -> ThreadWithReturnValue
        """Wrapping for the wgAdcCalib process
        """
        lib_adc_calibration = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgAdcCalib.so'))
        wg_adc_calibration = _wrap_function(lib_adc_calibration, 'wgAdcCalib', ct.c_int, [ct.c_char_p])
        env = WagasciEnvironment()
        adc_calibration_args = {
            "tree_files": tree_files if isinstance(tree_files, list) else [tree_files],
            "card_directory": env['WAGASCI_CONFDIR'],
            "topology_source": topology_source,
            "history_dir": history_dir,
            "history_file": history_file,
            "dif_id": int(dif_id),
            "pass": int(passes),
            "enable_plotting": bool(enable_plotting),
            "fixed_wallmrd_gain": bool(fixed_wallmrd_gain),
            "fixed_wallmrd_input_dac": int(fixed_wallmrd_input_dac),
            "silent_mode": bool(silent_mode)
        }
        return wg_adc_calibration(json.dumps(adc_calibration_args).encode('utf-8'))

    @wagascianpy.utils.utils.utf8_decorator
    def bcid_distribution(self, input_file, topology_source, output_img_dir, chip_by_chip):
        # type: (str, str, str, bool) -> ThreadWithReturnValue
        """Wrapping for the wgBCID process"""
        lib_bcid = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgBCID.so'))
        wg_bcid = _wrap_function(lib_bcid, 'wgBCID', ct.c_int, [ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_bool])
        return wg_bcid(input_file, topology_source, output_img_dir, chip_by_chip)

    @wagascianpy.utils.utils.utf8_decorator
    def adc_distribution(self, input_file, topology_source, output_img_dir, chip_by_chip):
        # type: (str, str, str, bool) -> ThreadWithReturnValue
        """Wrapping for the wgADC process"""
        lib_adc = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgADC.so'))
        wg_adc = _wrap_function(lib_adc, 'wgADC', ct.c_int, [ct.c_char_p, ct.c_char_p, ct.c_char_p, ct.c_bool])
        return wg_adc(input_file, topology_source, output_img_dir, chip_by_chip)

    @wagascianpy.utils.utils.utf8_decorator
    def tdc_calibration(self, input_file, config_file, dif):
        # type: (str, str, int) -> ThreadWithReturnValue
        """Wrapping for the wgTdcApply process"""
        lib_tdc = ct.CDLL(os.path.join(self.m_lib_dir, 'libwgTdcApply.so'))
        wg_tdc = _wrap_function(lib_tdc, 'wgTdcApply', ct.c_int, [ct.c_char_p, ct.c_char_p, ct.c_int])
        return wg_tdc(input_file, config_file, dif)

    @staticmethod
    def beam_summary_data(*args, **kwargs):
        try:
            wagascianpy.analysis.beam_summary_data.beam_summary_data(*args, **kwargs)
        except Exception as exception:
            print("beam_summary_data terminated with exception : %s" % str(exception))
            return wagascianpy.analysis.beam_summary_data.BSD_ERROR_CODE
        return 0

    @staticmethod
    def apply_detector_flags(*args, **kwargs):
        try:
            wagascianpy.analysis.apply_detector_flags.apply_detector_flags(*args, **kwargs)
        except Exception as exception:
            print("apply_detector_flags terminated with exception : %s" % str(exception))
            return wagascianpy.analysis.apply_detector_flags.DETECTOR_FLAGS_ERROR_CODE
        return 0

    @staticmethod
    def sanity_check(*args, **kwargs):
        try:
            wagascianpy.analysis.sanity_check.sanity_check(*args, **kwargs)
        except Exception as exception:
            print("sanity_check terminated with exception : %s" % str(exception))
            return wagascianpy.analysis.sanity_check.SANITY_CHECK_ERROR_CODE
        return 0

    @wagascianpy.utils.utils.utf8_decorator
    def get_dif_topology(self, acq_config_xml, dif_mapping_file=b"", mac_mapping_file=b""):
        # type: (str, str, str) -> Tuple[str, ct.c_void_p]
        """Get detector topology (map_dif) from the acquisition configuration xml
        file"""
        lib_wagasci = ct.CDLL(os.path.join(self.m_lib_dir, 'libwagasci.so'))
        wg_get_dif_topology = _wrap_function(lib_wagasci, 'GetDifTopologyCtypes', ct.c_void_p,
                                             [ct.c_char_p, ct.c_char_p, ct.c_char_p])
        pointer = wg_get_dif_topology(acq_config_xml, dif_mapping_file, mac_mapping_file)
        topology_string = ct.cast(pointer, ct.c_char_p).value
        if not isinstance(topology_string, str):
            topology_string = topology_string.decode('utf-8')
        print("Topology string allocated address (Python): %s" % hex(pointer))
        if topology_string == "":
            self.free_topology(pointer)
            raise ValueError("Received empty topology string")
        return topology_string, pointer

    @wagascianpy.utils.utils.utf8_decorator
    def get_gdcc_topology(self, acq_config_xml, dif_mapping_file=b"", mac_mapping_file=b""):
        # type: (str, str, str) -> Tuple[str, ct.c_void_p]
        """Get detector topology (map_gdcc) from the acquisition configuration xml
        file"""
        lib_wagasci = ct.CDLL(os.path.join(self.m_lib_dir, 'libwagasci.so'))
        wg_get_gdcc_topology = _wrap_function(lib_wagasci, 'GetGdccTopologyCtypes', ct.c_void_p,
                                              [ct.c_char_p, ct.c_char_p, ct.c_char_p])
        pointer = wg_get_gdcc_topology(acq_config_xml, dif_mapping_file, mac_mapping_file)
        topology_string = ct.cast(pointer, ct.c_char_p).value
        if not isinstance(topology_string, str):
            topology_string = topology_string.decode('utf-8')
        print("Topology string allocated address (Python): %s" % hex(pointer))
        if topology_string == "":
            self.free_topology(pointer)
            raise ValueError("Received empty topology string")
        return topology_string, pointer

    @wagascianpy.utils.utils.utf8_decorator
    def free_topology(self, pointer):
        """Release memory for the topology string"""
        lib_wagasci = ct.CDLL(os.path.join(self.m_lib_dir, 'libwagasci.so'))
        wg_free_topology = _wrap_function(lib_wagasci, 'FreeTopologyCtypes', None, [ct.c_void_p])
        print("Freeing topology string address (Python): %s" % hex(pointer))
        wg_free_topology(pointer)

    def enable_thread_safety(self):
        # type: (...) -> None
        """Enable ROOT thread safety"""
        lib_wagasci = ct.CDLL(os.path.join(self.m_lib_dir, 'libwagasci.so'))
        wg_enable_thread_safety = _wrap_function(lib_wagasci, 'wgEnableThreadSafety', None, None)
        wg_enable_thread_safety()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio
#

import json

from typing import Type, List

import wagascianpy.database.bsddb
import wagascianpy.database.wagascidb
import wagascianpy.utils.configuration
from wagascianpy.viewer.import_tkinter import import_tkinter, import_pygubu

pygubu = import_pygubu()


# noinspection PyArgumentList, PyUnresolvedReferences
def _show_run_info(builder,  # type: pygubu.Builder
                   record,  # type: wagascianpy.database.wagascidb.WagasciRunRecord
                   bsd_database  # type: wagascianpy.database.bsddb.BsdDatabase
                   ):
    # type: (...) -> None
    """Show info about a run record in the run_info frame
    """

    tkinter = import_tkinter()
    if not tkinter:
        print("Run info could not be should because tkinter module was not found")
        return

    run_info = builder.get_object('run_info')
    for child in run_info.winfo_children():
        child.destroy()

    pot_upper_limit = 0
    pot_lower_limit = 0
    good_bsd_spills_upper_limit = 0
    good_bsd_spills_lower_limit = 0
    neutrino_global_type = wagascianpy.database.bsddb.NeutrinosType.Unknown.name
    if bsd_database:
        try:
            bsd_records = bsd_database.get_time_interval(record["start_time"], record["stop_time"], True, True)
            if bsd_records:
                for bsd_record in bsd_records:
                    pot_upper_limit += int(bsd_record["number_of_pot"])
                    good_bsd_spills_upper_limit += int(bsd_record["number_of_good_spills"])
            bsd_records = bsd_database.get_time_interval(record["start_time"], record["stop_time"], True, False)
            if bsd_records:
                for bsd_record in bsd_records:
                    neutrino_type = wagascianpy.database.bsddb.NeutrinosType(bsd_records[0]["neutrino_type"])
                    if neutrino_type != wagascianpy.database.bsddb.NeutrinosType.Unknown:
                        neutrino_global_type = neutrino_type.name
                    pot_lower_limit += int(bsd_record["number_of_pot"])
                    good_bsd_spills_lower_limit += int(bsd_record["number_of_good_spills"])
        except (RuntimeWarning, ValueError) as exception:
            print(str(exception))

    # Run name
    label = tkinter.Label(run_info, text='Run name', padx=5)
    label.grid(row=0, column=0, sticky=tkinter.W)
    label = tkinter.Label(run_info, text=record['name'], padx=5)
    label.grid(row=0, column=1, sticky=tkinter.W)
    # Run number
    label = tkinter.Label(run_info, text='Run number', padx=5)
    label.grid(row=1, column=0, sticky=tkinter.W)
    label = tkinter.Label(run_info, text=record['run_number'], padx=5)
    label.grid(row=1, column=1, sticky=tkinter.W)
    # Run type
    label = tkinter.Label(run_info, text='Run type', padx=5)
    label.grid(row=2, column=0, sticky=tkinter.W)
    label = tkinter.Label(run_info, text=record['run_type'], padx=5)
    label.grid(row=2, column=1, sticky=tkinter.W)
    # Is bad
    label = tkinter.Label(run_info, text='Good run flag (good = 1, bad = 0)', padx=5)
    label.grid(row=3, column=0, sticky=tkinter.W)
    label = tkinter.Label(run_info, text=record['good_run_flag'], padx=5)
    label.grid(row=3, column=1, sticky=tkinter.W)
    # WAGASCI upstream good data flag
    label = tkinter.Label(run_info, text='WAGASCI upstream good data flag (good = 1, bad = 0)', padx=5)
    label.grid(row=4, column=0, sticky=tkinter.W)
    label = tkinter.Label(run_info, text=record['wagasci_upstream_good_data_flag'], padx=5)
    label.grid(row=4, column=1, sticky=tkinter.W)
    # WAGASCI downstream good data flag
    label = tkinter.Label(run_info, text='WAGASCI downstream good data flag (good = 1, bad = 0)', padx=5)
    label.grid(row=5, column=0, sticky=tkinter.W)
    label = tkinter.Label(run_info, text=record['wagasci_downstream_good_data_flag'], padx=5)
    label.grid(row=5, column=1, sticky=tkinter.W)
    # WallMRD north good data flag
    label = tkinter.Label(run_info, text='WallMRD north good data flag (good = 1, bad = 0)', padx=5)
    label.grid(row=6, column=0, sticky=tkinter.W)
    label = tkinter.Label(run_info, text=record['wallmrd_north_good_data_flag'], padx=5)
    label.grid(row=6, column=1, sticky=tkinter.W)
    # WallMRD south good data flag
    label = tkinter.Label(run_info, text='WallMRD south good data flag (good = 1, bad = 0)', padx=5)
    label.grid(row=7, column=0, sticky=tkinter.W)
    label = tkinter.Label(run_info, text=record['wallmrd_south_good_data_flag'], padx=5)
    label.grid(row=7, column=1, sticky=tkinter.W)
    # Duration
    label = tkinter.Label(run_info, text='Duration (hours)', padx=5)
    label.grid(row=8, column=0, sticky=tkinter.W)
    duration_h = "{0:.2f}".format(float(record['duration_h']), padx=5)
    label = tkinter.Label(run_info, text=duration_h, padx=5)
    label.grid(row=8, column=1, sticky=tkinter.W)
    # Start time
    label = tkinter.Label(run_info, text='Start time', padx=5)
    label.grid(row=9, column=0, sticky=tkinter.W)
    start_time = record.get_start_datetime().strftime('%Y/%m/%d %H:%M:%S %Z')
    label = tkinter.Label(run_info, text=start_time, padx=5)
    label.grid(row=9, column=1, sticky=tkinter.W)
    # Stop time
    label = tkinter.Label(run_info, text='Stop time', padx=5)
    label.grid(row=10, column=0, sticky=tkinter.W)
    stop_time = record.get_stop_datetime().strftime('%Y/%m/%d %H:%M:%S %Z')
    label = tkinter.Label(run_info, text=stop_time, padx=5)
    label.grid(row=10, column=1, sticky=tkinter.W)
    # POTs upper limit
    label = tkinter.Label(run_info, text='POTs (upper limit)', padx=5)
    label.grid(row=11, column=0, sticky=tkinter.W)
    label = tkinter.Label(run_info, text="{:.2e}".format(pot_upper_limit), padx=5)
    label.grid(row=11, column=1, sticky=tkinter.W)
    # POTs lower limit
    label = tkinter.Label(run_info, text='POTs (lower limit)', padx=5)
    label.grid(row=12, column=0, sticky=tkinter.W)
    label = tkinter.Label(run_info, text="{:.2e}".format(pot_lower_limit), padx=5)
    label.grid(row=12, column=1, sticky=tkinter.W)
    # Good BSD spills upper limit
    label = tkinter.Label(run_info, text='Good BSD spills (upper limit)', padx=5)
    label.grid(row=13, column=0, sticky=tkinter.W)
    label = tkinter.Label(run_info, text=str(good_bsd_spills_upper_limit), padx=5)
    label.grid(row=13, column=1, sticky=tkinter.W)
    # Good BSD spills lower limit
    label = tkinter.Label(run_info, text='Good BSD spills (lower limit)', padx=5)
    label.grid(row=14, column=0, sticky=tkinter.W)
    label = tkinter.Label(run_info, text=str(good_bsd_spills_lower_limit), padx=5)
    label.grid(row=14, column=1, sticky=tkinter.W)
    # Neutrino Type
    label = tkinter.Label(run_info, text='Neutrino type', padx=5)
    label.grid(row=15, column=0, sticky=tkinter.W)
    label = tkinter.Label(run_info, text=neutrino_global_type, padx=5)
    label.grid(row=15, column=1, sticky=tkinter.W)
    # Run folder
    label = tkinter.Label(run_info, text='Run folder', padx=5)
    label.grid(row=16, column=0, sticky=tkinter.W)
    label = tkinter.Label(run_info, text=record["run_folder"], padx=5)
    label.grid(row=16, column=1, sticky=tkinter.W)
    # Topology
    label = tkinter.Label(run_info, text='Topology', padx=5)
    label.grid(row=17, column=0, sticky=tkinter.W)
    if record["topology"] in ["", "undef", None]:
        pass
    else:
        topology_dict = json.loads(record['topology'])
        topology_str = ""
        for detector_name, detector_topology in wagascianpy.database.wagascidb.DETECTOR_TOPOLOGY.items():
            found = True
            for dif_id, dif_topology in detector_topology.items():
                if dif_id not in topology_dict or \
                        sorted(topology_dict[dif_id]) != sorted(dif_topology):
                    found = False
            if found:
                topology_str += detector_name + ", "
                label = tkinter.Label(run_info, text=topology_str.rstrip(", "), padx=5, wraplength=1000)
                label.grid(row=17, column=1, sticky=tkinter.W)
                # Raw files
    for i, (dif, raw_file) in enumerate(sorted(record['raw_files'].items()), 18):
        label = tkinter.Label(run_info, text="DIF %s" % dif, padx=5)
        label.grid(row=i, column=0, sticky=tkinter.W)
        label = tkinter.Label(run_info, text=raw_file, padx=5)
        label.grid(row=i, column=1, sticky=tkinter.W)


# noinspection PyUnresolvedReferences
def make_records_table(config,  # type: Type[wagascianpy.utils.configuration.Configuration]
                       builder,  # type: pygubu.Builder
                       records  # type: List[wagascianpy.database.wagascidb.WagasciRunRecord]
                       ):
    # type: (...) -> None

    tkinter = import_tkinter()
    if not tkinter:
        print("Run info could not be should because tkinter module was not found")
        return

    run_list = builder.get_object('run_list')
    for child in run_list.innerframe.winfo_children():
        child.destroy()
    run_info = builder.get_object('run_info')
    for run_child in run_info.winfo_children():
        run_child.destroy()

    label = tkinter.Label(run_list.innerframe, text='Run name')
    label.grid(row=0, column=0)
    label = tkinter.Label(run_list.innerframe, text='Run number')
    label.grid(row=0, column=1)
    label = tkinter.Label(run_list.innerframe, text='Run type')
    label.grid(row=0, column=2)
    label = tkinter.Label(run_list.innerframe, text='Duration (hours)')
    label.grid(row=0, column=3)
    label = tkinter.Label(run_list.innerframe, text='Start time')
    label.grid(row=0, column=4)
    label = tkinter.Label(run_list.innerframe, text='Stop time')
    label.grid(row=0, column=5)

    try:
        bsd_database = wagascianpy.database.bsddb.BsdDataBase(
            bsd_database_location=config.bsddb.bsd_database(),
            bsd_repository_location=config.bsddb.bsd_repository(),
            bsd_repository_download_location=config.bsddb.bsd_download_location(),
            t2kruns=config.global_configuration.t2krun(),
            update_db=config.viewer.update_bsd_database(),
            rebuild_db=config.viewer.rebuild_bsd_database())
    except RuntimeWarning as exception:
        bsd_database = None
        print(str(exception))

    records.sort(key=lambda x: int(x['run_number']))
    for i, record in enumerate(records, 1):
        button = tkinter.Button(run_list.innerframe, text=record['name'],
                                command=lambda rec=record, db=bsd_database: _show_run_info(builder, rec, db))
        button.grid(row=i, column=0)
        label = tkinter.Label(run_list.innerframe, text=record['run_number'])
        label.grid(row=i, column=1)
        label = tkinter.Label(run_list.innerframe, text=record['run_type'])
        label.grid(row=i, column=2)
        duration_h = "{0:.2f}".format(float(record['duration_h']))
        label = tkinter.Label(run_list.innerframe, text=duration_h)
        label.grid(row=i, column=3)
        start_time = record.get_start_datetime().strftime('%Y/%m/%d %H:%M:%S %Z')
        label = tkinter.Label(run_list.innerframe, text=start_time)
        label.grid(row=i, column=4, padx=5)
        stop_time = record.get_stop_datetime().strftime('%Y/%m/%d %H:%M:%S %Z')
        label = tkinter.Label(run_list.innerframe, text=stop_time)
        label.grid(row=i, column=5, padx=5)

    run_list.reposition()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

import abc
from collections import namedtuple

from six import string_types

import wagascianpy.database.wagascidb as wgdb
import wagascianpy.plotting.colors

import ROOT

ROOT.PyConfig.IgnoreCommandLineOptions = True

# compatible with Python 2 *and* 3
try:
    # noinspection PyUnresolvedReferences
    IntTypes = (int, long)  # Python2
except NameError:
    IntTypes = int  # Python3
ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})

CustomEvent = namedtuple('CustomEvent', ['start', 'stop', 'duration', 'label'])
MarkerTuple = namedtuple('MarkerTuple', ['run', 'maintenance', 'trouble', 'bsd'])


class Marker(object):

    def __init__(self, left_position=None, left_text=None,
                 line_color=wagascianpy.plotting.colors.Colors.Black.value):
        self._line_color = None
        self._left_text = None
        self._left_position = None
        self.line_color = line_color
        self.left_text = left_text
        self.left_position = left_position

    @property
    def left_text(self):
        return self._left_text

    @left_text.setter
    def left_text(self, left_text):
        self._left_text = left_text

    @property
    def left_position(self):
        return self._left_position

    @left_position.setter
    def left_position(self, left_position):
        self._left_position = left_position

    @property
    def line_color(self):
        return self._line_color

    @line_color.setter
    def line_color(self, line_color):
        if isinstance(line_color, IntTypes):
            self._line_color = line_color
        elif isinstance(line_color, wagascianpy.plotting.colors.Colors):
            self._line_color = line_color.value
        elif isinstance(line_color, string_types):
            self._line_color = wagascianpy.plotting.line_colors.Colors.get_by_detector(line_color).value
        else:
            raise TypeError("line color type is unknown {}".format(type(line_color).__name__))

    def make_tobjects(self):
        left_markers = []
        ROOT.gPad.Modified()
        ROOT.gPad.Update()
        y_minimum = ROOT.gPad.GetUymin()
        y_maximum = ROOT.gPad.GetUymax()
        x_minimum = ROOT.gPad.GetUxmin()
        x_maximum = ROOT.gPad.GetUxmax()
        tline = ROOT.TLine(self.left_position, y_minimum, self.left_position, y_maximum)
        tline.SetLineColor(self.line_color)
        tline.SetLineWidth(1)
        tline.SetLineStyle(2)
        if self.left_text:
            ttext = ROOT.TText(self.left_position + (x_maximum - x_minimum) * 0.005,
                               0.98 * y_maximum, self.left_text)
            ttext.SetTextAngle(90)
            ttext.SetTextAlign(33)
            ttext.SetTextSize(0.03)
            ttext.SetTextColor(self.line_color)
            left_markers.append(ttext)
        left_markers.append(tline)
        return left_markers


class DoubleMarker(Marker):

    def __init__(self, left_position=None, left_text="",
                 line_color=wagascianpy.plotting.colors.Colors.Black.value,
                 right_position=None, right_text=""):
        super(DoubleMarker, self).__init__(left_position=left_position,
                                           left_text=left_text, line_color=line_color)
        self._right_text = None
        self._right_position = None
        self.right_text = right_text
        self.right_position = right_position
        self._transparency = None
        self._fill_color = None

    @property
    def right_text(self):
        return self._right_text

    @right_text.setter
    def right_text(self, right_text):
        self._right_text = right_text

    @property
    def right_position(self):
        return self._right_position

    @right_position.setter
    def right_position(self, right_position):
        self._right_position = right_position

    @property
    def fill_color(self):
        return self._fill_color

    @fill_color.setter
    def fill_color(self, fill_color):
        if isinstance(fill_color, IntTypes):
            self._fill_color = fill_color
        elif isinstance(fill_color, wagascianpy.plotting.colors.Colors):
            self._fill_color = fill_color.value
        elif isinstance(fill_color, string_types):
            self._fill_color = wagascianpy.plotting.fill_colors.Colors.get_by_detector(fill_color).value
        else:
            raise TypeError("fill color type is unknown {}".format(type(fill_color).__name__))

    @property
    def transparency(self):
        return self._transparency

    @transparency.setter
    def transparency(self, transparency):
        if not 0 < transparency < 1:
            raise ValueError("Transparence must be a percentage from zero to one")
        self._transparency = transparency

    def make_tobjects(self):
        left_markers = super(DoubleMarker, self).make_tobjects()
        right_markers = []
        y_minimum = ROOT.gPad.GetUymin()
        y_maximum = ROOT.gPad.GetUymax()
        x_minimum = ROOT.gPad.GetUxmin()
        x_maximum = ROOT.gPad.GetUxmax()
        tline = ROOT.TLine(self.right_position, y_minimum, self.right_position, y_maximum)
        tline.SetLineColor(self.line_color)
        tline.SetLineWidth(1)
        tline.SetLineStyle(2)
        if self.fill_color:
            tbox = ROOT.TBox(self.left_position, y_minimum, self.right_position, y_maximum)
            tbox.SetLineWidth(0)
            tbox.SetFillColorAlpha(self.fill_color + 1, self.transparency)
            right_markers.append(tbox)
        if self.right_text:
            ttext = ROOT.TText(self.right_position - (x_maximum - x_minimum) * 0.025,
                               0.98 * y_maximum, self.right_text)
            ttext.SetTextAngle(90)
            ttext.SetTextAlign(33)
            ttext.SetTextSize(0.03)
            ttext.SetTextColor(self.line_color)
            right_markers.append(ttext)
        right_markers.append(tline)
        return right_markers + left_markers


class CustomEvents(ABC):
    _custom_events = []
    _fill_color = ROOT.kGray

    def __init__(self, start, stop=None, wagasci_database=None):
        # type: (Optional[Union[str, int]], Optional[Union[str, int]], Optional[str]) -> None
        self._start = start  # type: Optional[Union[str, int]]
        self._stop = stop  # type: Optional[Union[str, int]]
        self._wagasci_database = wagasci_database if wagasci_database else None  # type: Optional[str]
        if self._stop and type(start) is not type(stop):
            raise ValueError("Class of start and stop are different")
        if isinstance(start, IntTypes):
            self._start, self._stop = wgdb.run_to_interval(start=start, stop=stop, database=self._wagasci_database)
        elif isinstance(start, string_types):
            self._start = wgdb.WagasciRunRecord.str2datetime(start)
            self._stop = wgdb.WagasciRunRecord.str2datetime(stop)

    def get_interval(self, include_overlapping=False):
        interval = []
        for custom_event in self._custom_events:
            start_time = wgdb.WagasciRunRecord.str2datetime(custom_event.start)
            stop_time = wgdb.WagasciRunRecord.str2datetime(custom_event.stop)
            if include_overlapping:
                if self._start < start_time < self._stop or self._start < stop_time < self._stop:
                    interval.append(custom_event)
            else:
                if self._start < start_time < self._stop and self._start < stop_time < self._stop:
                    interval.append(custom_event)
        return interval

    def get_markers(self, include_overlapping=False):
        days = self.get_interval(include_overlapping)
        markers = []
        for day in days:
            start_timestamp = wgdb.WagasciRunRecord.str2timestamp(day.start)
            stop_timestamp = wgdb.WagasciRunRecord.str2timestamp(day.stop)
            marker = DoubleMarker(left_position=start_timestamp, right_position=stop_timestamp,
                                  left_text="{} ({})".format(day.label, day.start))
            marker.line_color = wagascianpy.plotting.colors.Colors.Black
            marker.fill_color = self._fill_color
            marker.transparency = 0.2
            markers.append(marker)
        return markers


class MaintenanceDays(CustomEvents):
    _fill_color = wagascianpy.plotting.colors.Colors.Green
    _custom_events = [
        CustomEvent('2019/11/07 05:30:00', '2019/11/07 17:30:00', 12, 'half day maintenance'),
        CustomEvent('2019/11/13 09:00:00', '2019/11/14 09:00:00', 24, 'one day maintenance'),
        CustomEvent('2019/11/20 05:30:00', '2019/11/20 17:30:00', 12, 'half day maintenance'),
        CustomEvent('2019/11/27 09:00:00', '2019/11/28 09:00:00', 24, 'one day maintenance'),
        CustomEvent('2019/12/04 05:30:00', '2019/12/04 17:30:00', 12, 'half day maintenance'),
        CustomEvent('2019/12/11 09:00:00', '2019/12/12 09:00:00', 24, 'one day maintenance'),
        CustomEvent('2019/12/19 09:00:00', '2020/01/14 15:00:00', 630, 'end of year shutdown'),
        CustomEvent('2020/01/16 05:30:00', '2020/01/16 17:30:00', 12, 'half day maintenance'),
        CustomEvent('2020/01/22 09:00:00', '2020/01/23 09:00:00', 24, 'one day maintenance'),
        CustomEvent('2020/01/29 05:30:00', '2020/01/29 17:30:00', 12, 'half day maintenance'),
        CustomEvent('2020/02/05 09:00:00', '2020/02/06 09:00:00', 24, 'one day maintenance'),
        CustomEvent('2020/02/12 05:30:00', '2020/02/12 17:30:00', 12, 'half day maintenance'),
        CustomEvent('2021/03/10 05:30:00', '2021/03/10 17:30:00', 12, 'half day maintenance'),
        CustomEvent('2021/03/17 09:00:00', '2021/03/18 09:00:00', 24, 'one day maintenance'),
        CustomEvent('2021/03/19 20:00:00', '2021/04/08 15:00:00', 475, 'beam stop'),
        CustomEvent('2021/04/14 05:30:00', '2021/04/14 17:30:00', 12, 'half day maintenance'),
        CustomEvent('2021/04/21 09:00:00', '2021/04/22 09:00:00', 24, 'one day maintenance')
    ]


class TroubleEvents(CustomEvents):
    _fill_color = wagascianpy.plotting.colors.Colors.Red
    _custom_events = [
        CustomEvent('2020/02/11 12:52:32', '2020/02/11 22:30:32', 9.6333, 'WallMRD south data corruption'),
        CustomEvent('2020/02/02 10:47:03', '2020/02/02 18:51:20', 8.0666, 'WallMRD north data corruption'),
        CustomEvent('2020/02/04 09:04:39', '2020/02/04 14:07:09', 5.0416, 'WallMRD north data corruption'),
        CustomEvent('2020/01/31 09:56:07', '2020/01/31 14:35:42', 4.6597, 'WallMRD south data corruption'),
        CustomEvent('2020/01/30 22:19:00', '2020/01/30 22:48:00', 0.4833, 'WallMRD south HV failed'),
        CustomEvent('2020/01/30 12:52:00', '2020/01/30 13:48:00', 0.9333, 'WAGASCI downstream HV failed'),
        CustomEvent('2020/01/29 17:49:16', '2020/01/29 18:26:00', 0.6, 'WAGASCI run 91 : Data corruption'),
        CustomEvent('2020/01/24 16:09:35', '2020/01/24 16:26:44', 0.2833, 'WAGASCI run 82-83 : Data corruption'),
        CustomEvent('2020/01/21 11:02:00', '2020/01/21 15:45:00', 4.7166, 'WAGASCI run 73-77 : Data corruption'),
        CustomEvent('2020/01/20 19:51:24', '2020/01/21 14:22:05', 18.5, 'WAGASCI run 72 overwritten : human error'),
        # CustomEvent('2020/01/20 19:30:00', '2020/01/20 19:50:00', 0.3333, 'CCC trouble'),
        CustomEvent('2020/01/14 10:46:26', '2020/01/14 12:57:09', 1.1833, 'WAGASCI run 56-57 : Data corruption'),
        CustomEvent('2019/12/15 14:39:26', '2020/12/16 12:31:58', 21.866, 'WAGASCI run 43-48 : Data corruption'),
        CustomEvent('2019/11/26 04:03:02', '2019/11/27 09:00:00', 4.9494, 'WAGASCI run 26 : Data corruption'),
        CustomEvent('2019/11/18 14:51:19', '2019/11/18 18:22:16', 3.5158, 'WAGASCI downstream data corruption'),
        CustomEvent('2019/11/17 12:56:07', '2019/11/17 20:13:41', 7.2927, 'WAGASCI downstream data corruption'),
        CustomEvent('2019/11/11 07:04:45', '2019/11/11 11:39:58', 4.5869, 'WAGASCI upstream data corruption'),
        CustomEvent('2019/11/10 19:23:23', '2019/11/11 11:39:58', 16.2763, 'WAGASCI downstream data corruption'),
        CustomEvent('2021/03/13 15:55:07', '2021/03/13 16:19:17', 0.4028, 'Replaced chip in WAGASCI downstream'),
        CustomEvent('2021/03/14 12:08:47', '2021/03/15 08:57:25', 20.8105, 'WAGASCI run 140 : Pyrame crash')
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# copyright 2020 pintaudi giorgio
#

import argparse
import sys
import time

from wagascianpy.utils.configuration import *
import wagascianpy.utils.environment


def fill_configuration(args=None):
    # type: (Optional[[argparse.Namespace, Dict[str, Any]]]) -> None

    if args is None:
        args = {'batch_mode': False}

    if isinstance(args, argparse.Namespace):
        args = vars(args)

    batch_mode = args["batch_mode"]

    if batch_mode:
        # WAGASCI database configuration
        wagasci_database = args["wagasci_database"]
        wagasci_repository = args["wagasci_repository"]
        wagasci_download_location = args["wagasci_download_location"]
        wagasci_decoded_location = args["wagasci_decoded_location"]
        borg_wagasci_repository = args["borg_wagasci_repository"]
        simple_wagasci_repository = args["simple_wagasci_repository"]
        # BSD database configuration
        bsd_repository = args["bsd_repository"]
        bsd_database = args["bsd_database"]
        bsd_download_location = args["bsd_download_location"]
        # global configuration
        t2krun = args["t2krun"]
        history_location = args["history_location"]
        wagasci_libdir = args["wagasci_libdir"]
        # temperature configuration
        temperature_sqlite_database = args["temperature_sqlite_database"]
        # viewer volatile configuration
        update_wagasci_database = args["update_wagasci_database"]
        rebuild_wagasci_database = args["rebuild_wagasci_database"]
        update_bsd_database = args["update_bsd_database"]
        rebuild_bsd_database = args["rebuild_bsd_database"]
        only_good_runs = args["only_good_runs"]
        include_overlapping = args["include_overlapping"]
        # analyzers
        download = args["download"]
        decoder = args["decoder"]
        make_hist = args["make_hist"]
        tdc_calibration = args["tdc_calibration"]
        bcid_distribution = args["bcid_distribution"]
        adc_distribution = args["adc_distribution"]
        spill_number_fixer = args["spill_number_fixer"]
        beam_summary_data = args["beam_summary_data"]
        apply_detector_flags = args["apply_detector_flags"]
        sanity_check = args["sanity_check"]
        temperature = args["temperature"]
        adc_calibration = args["adc_calibration"]
        all_analyzers = args["all_analyzers"]
        overwrite_flag = args["overwrite_flag"]
        # run selectors
        start_time = args["start_time"]
        stop_time = args["stop_time"]
        start_run = args["start_run"]
        stop_run = args["stop_run"]
    else:
        # WAGASCI database configuration
        wagasci_database = None
        wagasci_repository = None
        wagasci_download_location = None
        wagasci_decoded_location = None
        borg_wagasci_repository = None
        simple_wagasci_repository = None
        # BSD database configuration
        bsd_repository = None
        bsd_database = None
        bsd_download_location = None
        # global configuration
        t2krun = None
        history_location = None
        wagasci_libdir = None
        # temperature configuration
        temperature_sqlite_database = None
        # viewer volatile configuration
        update_wagasci_database = None
        rebuild_wagasci_database = None
        update_bsd_database = None
        rebuild_bsd_database = None
        only_good_runs = None
        include_overlapping = None
        # analyzers
        download = None
        decoder = None
        make_hist = None
        tdc_calibration = None
        bcid_distribution = None
        adc_distribution = None
        spill_number_fixer = None
        beam_summary_data = None
        apply_detector_flags = None
        sanity_check = None
        temperature = None
        adc_calibration = None
        all_analyzers = None
        overwrite_flag = None
        # run selectors
        start_time = None
        stop_time = None
        start_run = None
        stop_run = None

    # sanity checks
    if int(bool(borg_wagasci_repository)) + int(bool(simple_wagasci_repository)) == 2:
        raise ValueError("cannot select both --borg-wagasci-repository and --simple-wagasci-repository")
    get_time_interval = False
    get_run_interval = False
    get_all = False
    if start_time:
        if not stop_time:
            raise ValueError("please select stop time using --stop-time")
        get_time_interval = True
    elif start_run:
        if not stop_run:
            stop_run = start_run
        get_run_interval = True
    else:
        get_all = True
    if all_analyzers:
        decoder = tdc_calibration = bcid_distribution = adc_distribution = spill_number_fixer = True
        beam_summary_data = temperature = adc_calibration = apply_detector_flags = sanity_check = True

    # WAGASCI library directory
    if not wagasci_libdir:
        try:
            env = wagascianpy.utils.environment.WagasciEnvironment()
            wagasci_libdir = env["WAGASCI_LIBDIR"]
        except KeyError:
            pass

    # wagasci database configuration
    if not wagasci_database:
        wagasci_database = Configuration.wagascidb.wagasci_database()
    if not wagasci_repository:
        wagasci_repository = Configuration.wagascidb.wagasci_repository()
    if not wagasci_download_location:
        wagasci_download_location = Configuration.wagascidb.wagasci_download_location()
    if not wagasci_decoded_location:
        wagasci_decoded_location = Configuration.wagascidb.wagasci_decoded_location()
    if borg_wagasci_repository:
        repository_type = RepositoryType.Borg
    elif simple_wagasci_repository:
        repository_type = RepositoryType.Simple
    else:
        repository_type = Configuration.wagascidb.repository_type()

    # bsd database configuration
    if not bsd_database:
        bsd_database = Configuration.bsddb.bsd_database()
    if not bsd_repository:
        bsd_repository = Configuration.bsddb.bsd_repository()
    if not bsd_download_location:
        bsd_download_location = Configuration.bsddb.bsd_download_location()

    # global configuration
    if not t2krun:
        t2krun = Configuration.global_configuration.t2krun()
    if not history_location:
        history_location = Configuration.global_configuration.history_location()
    if not wagasci_libdir:
        wagasci_libdir = Configuration.global_configuration.wagasci_libdir()

    # temperature configuration
    if not temperature_sqlite_database:
        temperature_sqlite_database = Configuration.temperature.temperature_sqlite_database()

    # WAGASCI database configuration
    Configuration.wagascidb.override({
        'wagasci_database': wagasci_database,
        'wagasci_repository': wagasci_repository,
        'wagasci_download_location': wagasci_download_location,
        'wagasci_decoded_location': wagasci_decoded_location,
        'repository_type': repository_type
    })

    # BSD database configuration
    Configuration.bsddb.override({
        'bsd_repository': bsd_repository,
        'bsd_database': bsd_database,
        'bsd_download_location': bsd_download_location
    })

    # global configuration
    Configuration.global_configuration.override({
        't2krun': t2krun,
        'history_location': history_location,
        'wagasci_libdir': wagasci_libdir
    })

    # temperature configuration
    Configuration.temperature.override({
        'temperature_sqlite_database': temperature_sqlite_database
    })

    # viewer configuration
    Configuration.create_section('viewer')
    Configuration.viewer.override({
        'batch_mode': batch_mode,
        'update_wagasci_database': update_wagasci_database,
        'rebuild_wagasci_database': rebuild_wagasci_database,
        'update_bsd_database': update_bsd_database,
        'rebuild_bsd_database': rebuild_bsd_database,
        'only_good_runs': only_good_runs,
        'include_overlapping': include_overlapping
    })

    # analyzers configuration
    Configuration.create_section('analyzer_configuration')
    Configuration.analyzer_configuration.override({
        'download': download,
        'decoder': decoder,
        'make_hist': make_hist,
        'tdc_calibration': tdc_calibration,
        'bcid_distribution': bcid_distribution,
        'adc_distribution': adc_distribution,
        'spill_number_fixer': spill_number_fixer,
        'beam_summary_data': beam_summary_data,
        'apply_detector_flags': apply_detector_flags,
        'sanity_check': sanity_check,
        'temperature': temperature,
        'adc_calibration': adc_calibration,
        'all_analyzers': all_analyzers,
        'overwrite_flag': overwrite_flag
    })

    # run selectors
    Configuration.create_section('run_select')
    Configuration.run_select.override({
        'get_time_interval': get_time_interval,
        'get_run_interval': get_run_interval,
        'get_all': get_all,
        'start_time': start_time,
        'stop_time': stop_time,
        'start_run': start_run,
        'stop_run': stop_run
    })

    print("CONFIGURATION\n")
    Configuration.dump()
    sys.stdout.flush()
    time.sleep(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2020 Pintaudi Giorgio

# Python modules
import operator
import os
import re
import sys
from collections import namedtuple
from itertools import islice
from typing import Tuple, List, Optional, Union, Iterable, Dict

import wagascianpy.analysis.spill
import wagascianpy.database.db_record
import wagascianpy.utils.treenames as treenames
import wagascianpy.utils.utils
from wagascianpy.analysis.spill import open_raw_tree
from wagascianpy.database.bsddb import BunchNumber, CTNumber

# ROOT
try:
    import ROOT

    ROOT.PyConfig.IgnoreCommandLineOptions = True
except ImportError:
    ROOT = None

BSD_ERROR_CODE = 1000

_CONVERTION_FACTOR = 0x7FFF
_OFFSET = 1
_SEARCH_WINDOW = 500
_BSD_INTERVAL_OFFSET = 3600  # seconds

MatchedSpill = namedtuple('MatchedSpill', ['WagasciSpill', 'BsdSpill'])


def _set_wagasci_spills(input_file, matched_spills):
    ttree, tfile = open_raw_tree(input_file=input_file,
                                 branch_list=["spill_count", "spill_mode", "fixed_spill_number", "good_spill_flag"],
                                 mode="UPDATE")
    if tfile.GetListOfKeys().Contains(treenames.BSD_TREE_NAME):
        for key in tfile.GetListOfKeys():
            if key.GetName() == treenames.BSD_TREE_NAME:
                tfile.Delete("{};{}".format(treenames.BSD_TREE_NAME, key.GetCycle()))

    friend_tree = ROOT.TTree(treenames.BSD_TREE_NAME, "ROOT tree containing Beam Summary Data related information")
    friend_tree.SetDirectory(tfile)
    ttree.AddFriend(friend_tree)

    empty_bsd_spill = wagascianpy.analysis.spill.SpillFactory.get_spill("bsd")
    array_list = empty_bsd_spill.get_array_list()
    for array_info in array_list:
        friend_tree.Branch(array_info.name, array_info.array, array_info.type_str)

    for event in ttree:
        current_spill_number = event.fixed_spill_number
        current_spill_count = event.spill_count
        current_spill_mode = event.spill_mode
        if current_spill_mode != wagascianpy.analysis.spill.WAGASCI_SPILL_BEAM_MODE:
            matched_spill = None
        else:
            matched_spill = next((matched_spill for matched_spill in matched_spills
                                  if matched_spill.WagasciSpill.spill_number == current_spill_number and
                                  matched_spill.WagasciSpill.spill_count == current_spill_count), None)
        if matched_spill is None:
            bsd_spill = wagascianpy.analysis.spill.SpillFactory.get_spill("bsd")
        else:
            bsd_spill = matched_spill.BsdSpill
        bsd_spill.set_array_list(array_list)
        friend_tree.Fill()

    ttree.SetBranchStatus("*", 1)
    friend_tree.Write("", ROOT.TObject.kWriteDelete)
    tfile.Write("", ROOT.TObject.kWriteDelete)
    tfile.Close()


def _get_wagasci_spills(input_file):
    # type: (str) -> Tuple[List[wagascianpy.analysis.spill.WagasciSpill], int, int]
    """
    Read all the WAGASCI spills contained in the input file into a list of WagasciSpill objects.
    Each element of the list correspond to a single spill.
    :param input_file: path to the input file
    :return: list of WagasciSpill, starting timestamp, stopping timestamp
    """

    ttree, tfile = open_raw_tree(input_file=input_file,
                                 branch_list=["spill_count", "spill_mode", "fixed_spill_number", "good_spill_flag"],
                                 mode="READ")

    assert (ttree.GetUserInfo().FindObject("start_time") not in [ROOT.nullptr, None, 0]), \
        "Start time not found in {} TTree of file {}".format(ttree.GetName(), input_file)
    start_time = ttree.GetUserInfo().FindObject("start_time").GetVal()
    assert (ttree.GetUserInfo().FindObject("stop_time") not in [ROOT.nullptr, None, 0]), \
        "stop time not found in {} TTree of file {}".format(ttree.GetName(), input_file)
    stop_time = ttree.GetUserInfo().FindObject("stop_time").GetVal()
    dif_id = ttree.GetUserInfo().FindObject("dif_id").GetVal()
    print("Run start time %s" % wagascianpy.database.db_record.DBRecord.timestamp2str(start_time))
    print("Run stop time %s" % wagascianpy.database.db_record.DBRecord.timestamp2str(stop_time))
    print("DIF %s" % dif_id)
    regex = re.compile(r'_([\d]+)_ecal_dif_\d+_tree.root')
    match = re.search(regex, input_file)
    run = match.group(1) if match else 0
    wagasci_spills = []
    for event in ttree:
        if event.spill_mode != wagascianpy.analysis.spill.WAGASCI_SPILL_BEAM_MODE:
            continue
        wagasci_spill = wagascianpy.analysis.spill.SpillFactory.get_spill("wagasci")
        wagasci_spill.spill_number = event.fixed_spill_number
        wagasci_spill.spill_count = event.spill_count
        wagasci_spill.converted_spill_number = wagasci_spill.spill_number & _CONVERTION_FACTOR
        wagasci_spill.spill_mode = event.spill_mode
        wagasci_spill.good_spill_flag = event.good_spill_flag
        wagasci_spill.wagasci_run = run
        if wagasci_spill.are_all_defined():
            wagasci_spills.append(wagasci_spill)

    tfile.Close()
    return wagasci_spills, start_time, stop_time


def _check_bsd_file(bsd_record, bsd_repository, file_path):
    # type: (Dict, str, str) -> None
    if not os.path.exists(file_path):
        if bsd_repository is None:
            raise RuntimeError("Please specify a valid BSD local repository")
        file_path = wagascianpy.utils.utils.find_first_match(bsd_record["name"], bsd_repository)
        if file_path is None:
            raise RuntimeError("Please specify a valid BSD local repository")


def get_bsd_spills(bsd_database, bsd_repository, t2krun, start_time, stop_time):
    # type: (str, str, int, int, int) -> List[wagascianpy.analysis.spill.BsdSpill]
    """
    Read all the BSD spills from the BSD repository in the selected time interval
    :param bsd_database: path to BSD database file
    :param bsd_repository: path to the BSD repository directory
    :param t2krun: number of T2K run
    :param start_time: starting timestamp
    :param stop_time: stopping timestamp
    :return:
    """
    bsd_spills = []
    timestamp_start = wagascianpy.database.db_record.DBRecord.datetime2timestamp(start_time) - _BSD_INTERVAL_OFFSET
    timestamp_stop = wagascianpy.database.db_record.DBRecord.datetime2timestamp(stop_time) + _BSD_INTERVAL_OFFSET
    with wagascianpy.database.bsddb.BsdDataBase(bsd_database, bsd_repository, None, t2krun) as db:
        bsd_records = db.get_time_interval(datetime_start=timestamp_start, datetime_stop=timestamp_stop,
                                           include_overlapping=False)
        if not bsd_records:
            RuntimeError("Please specify a valid BSD database")
        for bsd_record in sorted(bsd_records, key=operator.itemgetter("name")):
            file_path = bsd_record["file_path"]
            if not os.path.exists(file_path):
                file_path = os.path.join(bsd_repository, "t2krun{}".format(t2krun),
                                         os.path.basename(bsd_record["file_path"]))
                if not os.path.exists(file_path):
                    file_path = os.path.join(bsd_repository, os.path.basename(bsd_record["file_path"]))
                    if not os.path.exists(file_path):
                        raise RuntimeError(
                            "BSD file {} not found in repository {}".format(bsd_record["file_path"], bsd_repository))
            _check_bsd_file(bsd_record=bsd_record, bsd_repository=bsd_repository, file_path=file_path)
            print('Reading file "{}"'.format(file_path))
            bsd_file = ROOT.TFile(file_path, "READ")
            bsd_tree = bsd_file.Get(treenames.BSD_TREE_NAME)

            bsd_tree.SetBranchStatus("*", 0)
            bsd_tree.SetBranchStatus("spillnum", 1)
            bsd_tree.SetBranchStatus("trg_sec", 1)
            bsd_tree.SetBranchStatus("trg_nano", 1)
            bsd_tree.SetBranchStatus("good_spill_flag", 1)

            ct_pot = bsd_tree.GetListOfBranches().FindObject("ct_pot")
            if ct_pot:
                bsd_tree.SetBranchStatus("ct_pot", 1)
            ct_np = bsd_tree.GetListOfBranches().FindObject("ct_np")
            if ct_np:
                bsd_tree.SetBranchStatus("ct_np", 1)

            for event in bsd_tree:
                # print("Spill number = %s" % event.spillnum)
                # print("POT = %s" % event.ct_pot[BunchNumber.TotalCurrent])
                # print("Is bad spill = %s" % event.good_spill_flag)
                # print("Timestamp = %s" % bsd_tree.trg_sec[TriggerTime.RubidiumClock]

                bsd_spill = wagascianpy.analysis.spill.SpillFactory.get_spill("bsd")
                bsd_spill.bsd_spill_number = int(event.spillnum)
                bsd_spill.converted_spill_number = (bsd_spill.bsd_spill_number + _OFFSET) & _CONVERTION_FACTOR
                bsd_spill.timestamp = float(event.trg_sec[wagascianpy.database.bsddb.TriggerTime.RubidiumClock])
                bsd_spill.timestamp += float(event.trg_nano[wagascianpy.database.bsddb.TriggerTime.RubidiumClock]) / 1e9
                bsd_spill.bsd_good_spill_flag = event.good_spill_flag
                bsd_spill.t2k_run = t2krun
                bsd_spill.main_ring_run = bsd_record.main_ring_run
                bsd_spill.neutrino_daq_run = bsd_record.neutrino_daq_run
                bsd_spill.horn_current = bsd_record.mean_horn_current  # TODO
                bsd_spill.neutrino_mode = bsd_record.neutrino_type

                for i in [i for i in BunchNumber if i != BunchNumber.TotalCurrent]:
                    if ct_pot:
                        bunch_pot = float(event.ct_pot[i])
                    elif ct_np:
                        bunch_pot = float(event.ct_np[CTNumber.Fifth * len(BunchNumber) + i])
                    else:
                        bunch_pot = 0
                    bsd_spill.bunch_pot[i - BunchNumber.First] = bunch_pot
                if ct_pot:
                    bsd_spill.pot = float(event.ct_pot[BunchNumber.TotalCurrent])
                elif ct_np:
                    bsd_spill.pot = float(event.ct_np[CTNumber.Fifth * len(BunchNumber) + BunchNumber.TotalCurrent])
                else:
                    bsd_spill.pot = 0

                if bsd_spill.are_all_defined():
                    bsd_spills.append(bsd_spill)
                else:
                    raise ValueError("Some fields of BSD spill are not initialized")

    return bsd_spills


def _match_spills(wagasci_spills,  # type: List[wagascianpy.analysis.spill.WagasciSpill]
                  bsd_spills,  # type: List[wagascianpy.analysis.spill.BsdSpill]
                  start_time,  # type: int
                  stop_time  # type: int
                  ):
    # type: (...) -> List[MatchedSpill]
    """
    Match the WAGASCI spill list with the BSD spill list and put the matched spills in a new list
    :param wagasci_spills: list of WAGASCI spills
    :param bsd_spills: list of BSD spills
    :param start_time: start timestamp of the lists
    :param stop_time: stop timestamp of the lists
    :return: list of matched spills
    """

    matched_spills = []

    if not wagasci_spills:
        raise ValueError("WAGASCI spill list is empty")
    if not bsd_spills:
        return matched_spills

    if bsd_spills[0].timestamp > start_time:
        ignore_spills = True
        print("[WARNING] first BSD spill %s is after run start %s" % (bsd_spills[0].timestamp, start_time))
    else:
        ignore_spills = False
    if bsd_spills[-1].timestamp < stop_time:
        print("[WARNING] last BSD spill %s is before run stop %s" % (bsd_spills[-1].timestamp, stop_time))

    good_spills = [wagasci_spill for wagasci_spill in wagasci_spills
                   if wagasci_spill.good_spill_flag == wagascianpy.analysis.spill.IS_GOOD_SPILL and
                   wagasci_spill.spill_mode == wagascianpy.analysis.spill.WAGASCI_SPILL_BEAM_MODE]

    found_index = None
    start_index = 0

    if ignore_spills:
        while good_spills:
            if abs(good_spills[0].converted_spill_number - bsd_spills[0].converted_spill_number) > 100:
                good_spills.pop(0)
            else:
                break

    for i, wagasci_spill in enumerate(good_spills):
        if i != 0 and i % 1000 == 0:
            print("Processed {} spills".format(i))
        # In case the previous spill was not found, increase the search window by a factor 10
        start_slice = int(max(start_index - _SEARCH_WINDOW / 2, 0))
        if found_index is None:
            stop_slice = int(min(start_slice + 10 * _SEARCH_WINDOW, sys.maxsize))
        else:
            stop_slice = int(min(start_slice + _SEARCH_WINDOW, sys.maxsize))

        # If the spill is not found, found_index and bsd_spill are set to None
        # Keep in mind that found_index is relative to the start of the slice and not to the start of bsd_spills
        found_index, bsd_spill = _find_spill(wagasci_spill.converted_spill_number,
                                             islice(bsd_spills, start_slice, stop_slice))

        # If the spill was found, set the start index as the found index
        if found_index is not None:
            start_index = int(start_slice + found_index)
            matched_spill = MatchedSpill(wagasci_spill, bsd_spill)
            matched_spills.append(matched_spill)
        else:
            start_index += 1
    return matched_spills


###############################################################################
#                                 find_spill                                  #
###############################################################################

def _find_spill(converted_spill_number,  # type: int
                spills  # type: Union[islice, Iterable[wagascianpy.analysis.spill.BsdSpill]]
                ):
    # type: (...) -> Tuple[Optional[int], Optional[wagascianpy.analysis.spill.BsdSpill]]
    """
    Find a spill with converted spill number in the input list of spills. Returns None in case of failure
    :param converted_spill_number: converted spill number (using the convertion flag)
    :param spills: slice of BSD spills
    :return: index of found spill, found BSD spill with matching converted spill number
    """
    for index, spill in enumerate(spills):
        if spill.converted_spill_number == converted_spill_number:
            return index, spill
    return None, None


###############################################################################
#                             beam_summary_data                               #
###############################################################################

def beam_summary_data(input_path, bsd_database, bsd_repository, t2krun, recursive):
    # type: (str, str, str, int, bool) -> None
    """
    Integrate the beam summary data (BSD) into the WAGASCI decoded data
    :param input_path: path to the tree.root file containing the WAGASCI decoded data or
                       a directory containing those files
    :param bsd_database: path to BSD database file .db
    :param bsd_repository: path to the BSD repository (where the BSD .root files are stored)
    :param t2krun: number of T2K run
    :param recursive: if True all the tree.root files in the input directory are analyzed
    :return: None
    """
    if ROOT is None:
        raise ImportError("No ROOT module was found!")
    if not os.path.exists(input_path):
        raise IOError("Input file %s not found" % input_path)
    input_files = []
    if os.path.isfile(input_path):
        input_files.append(input_path)
    elif os.path.isdir(input_path):
        if recursive:
            for root, dirs, files in os.walk(input_path):
                for file_name in files:
                    if re.search(r'_ecal_dif_([\d]+)_tree.root$', file_name) is not None:
                        input_files.append(os.path.join(root, file_name))
        else:
            for file_name in os.listdir(input_path):
                if re.search(r'_ecal_dif_([\d]+)_tree.root', file_name).groups() is not None:
                    input_files.append(os.path.join(input_path, file_name))

    if not os.path.exists(bsd_database):
        raise IOError("Beam summary data database %s not found" % bsd_database)

    for input_file in input_files:
        print("Start reading WAGASCI spills")
        wagasci_spills, start_time, stop_time = _get_wagasci_spills(input_file)
        print("Finished reading WAGASCI spills")
        print("Start reading BSD spills")
        bsd_spills = get_bsd_spills(bsd_database, bsd_repository, t2krun, start_time, stop_time)
        print("Finished reading BSD spills")
        print("Start WAGASCI-BSD spill matching")
        matched_spills = _match_spills(wagasci_spills, bsd_spills, start_time, stop_time)
        print("Finished WAGASCI-BSD spill matching")
        print("Start writing matched spills")
        _set_wagasci_spills(input_file, matched_spills)
        print("Finished writing matched spills")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2019 Pintaudi Giorgio, Eguchi Aoi
import os
import shutil

from six import string_types

import wagascianpy.analysis.analysis
import wagascianpy.utils.utils
from wagascianpy.utils.environment import WagasciEnvironment

_CHANGE_INPUTDAC_MODE = 2
_CHANGE_THRESHOLD_MODE = 0
_CHANGE_CHIPID_MODE = 5


###############################################################################
#                                change_bitstream                                #
###############################################################################

def change_bitstream(dif_topology, bitstream_root_dir, input_dac, trigger_threshold):
    """
    Change the global input DAC and global trigger threshold bitstream
    parameters for all the chips.

    Args:
        dif_topology: DIF topology dictionary
        bitstream_root_dir: root directory for the bistream folder tree
        input_dac: global input DAC value (0-255)
        trigger_threshold: global trigger threshold (0-1023)

    Returns:
        None

    Exceptions:
        ValueError if the arguments are empty or invalid
    """
    # Check sanity of arguments

    if not isinstance(dif_topology, dict):
        raise ValueError("[change_bitstream] DIF topology is not a dictionary")
    if not dif_topology:
        raise ValueError("[change_bitstream] DIF topology is empty")
    if not isinstance(bitstream_root_dir, string_types):
        raise ValueError("[change_bitstream] bitstream_dir is not a directory")
    if not bitstream_root_dir:
        raise ValueError("[change_bitstream] bitstream_dir is empty")
    bitstream_root_dir = bitstream_root_dir.rstrip("/")
    wagascianpy.utils.utils.mkdir_p(bitstream_root_dir)
    input_dac = int(input_dac)
    if input_dac < 0 or input_dac > 255:
        raise ValueError("[change_bitstream] input DAC value is not valid : %d"
                         % input_dac)
    trigger_threshold = int(trigger_threshold)
    if trigger_threshold < 0 or trigger_threshold > 1023:
        raise ValueError("[change_bitstream] trigger threshold value is not valid : %d"
                         % trigger_threshold)

    environ = WagasciEnvironment()
    try:
        bitstream_template = os.path.join(environ['WAGASCI_CONFDIR'], environ['WAGASCI_BITSTREAM_TEMPLATE'])
    except KeyError as error:
        raise KeyError("Could not find WAGASCI configuration directory : {}".format(str(error)))
    try:
        process = wagascianpy.analysis.analysis.WagasciAnalysis(environ['WAGASCI_LIBDIR'])
    except KeyError as error:
        raise KeyError("Could not find WAGASCI library directory : {}".format(str(error)))
    overwrite_flag = True
    edit_flag = True

    for dif, chip_map in sorted(iter(dif_topology.items()), key=lambda item: int(item[0])):
        for chip, n_channels in sorted(iter(chip_map.items()), key=lambda item: int(item[0])):
            chip_bitstream = os.path.join(bitstream_root_dir, "wagasci_bitstream_dif%s_chip%02d.txt" % (dif, int(chip)))
            if not os.path.isfile(chip_bitstream):
                shutil.copyfile(bitstream_template, chip_bitstream)
            # Input DAC
            for channel in range(wagascianpy.analysis.analysis.SPIROC2D_NCHANNELS):
                input_dac_value = input_dac
                if channel >= int(n_channels):
                    input_dac_value = -1
                result = process.change_config(chip_bitstream,
                                               chip_bitstream,
                                               overwrite_flag,
                                               edit_flag,
                                               input_dac_value,
                                               _CHANGE_INPUTDAC_MODE,
                                               channel)
                if result != 0:
                    raise ValueError("[change_bitstream] Error while changing "
                                     "input DAC : %d" % result)
            # Threshold
            result = process.change_config(chip_bitstream,
                                           chip_bitstream,
                                           overwrite_flag,
                                           edit_flag,
                                           trigger_threshold,
                                           _CHANGE_THRESHOLD_MODE)
            if result != 0:
                raise ValueError("[change_bitstream] Error while changing "
                                 "global threshold : %d" % result)
            # Chip ID
            result = process.change_config(chip_bitstream,
                                           chip_bitstream,
                                           overwrite_flag,
                                           edit_flag,
                                           chip,
                                           _CHANGE_CHIPID_MODE)
            if result != 0:
                raise ValueError("[change_bitstream] Error while changing "
                                 "Chip ID : %d" % result)
